# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions needed for XManager API implementation.

This module is private and can only be used by the API itself, but not by users.
"""

import abc
import asyncio
import enum
import functools
import os
import shlex
import sys
from typing import Any, Callable, TypeVar

from absl import flags
import attr
from xmanager.xm import pattern_matching

FLAGS = flags.FLAGS
flags.DEFINE_string(
   'xm_launch_script', None, 'Path to the launch script that is using '
   'XManager Launch API')

ReturnType = TypeVar('ReturnType')


class SpecialArg(abc.ABC):
  """A base class for arguments with special handling on serialization."""


@attr.s(auto_attribs=True)
class ShellSafeArg(SpecialArg):
  """Command line argument that shouldn't be escaped.

  Normally all arguments would be passed to the binary as is. To let shell
  substitutions (such as environment variable expansion) to happen the argument
  must be wrapped with this structure.
  """

  arg: str


ARG_ESCAPER = pattern_matching.match(
    pattern_matching.Case([ShellSafeArg], lambda v: v.arg),
    pattern_matching.Case([enum.Enum], lambda v: shlex.quote(str(v.name))),
    pattern_matching.Case([Any], lambda v: shlex.quote(str(v))),
)


def run_in_asyncio_loop(
    f: Callable[..., ReturnType]) -> Callable[..., ReturnType]:
  """A decorator that turns an async function to a synchronous one.

  Python asynchronous APIs can't be used directly from synchronous functions.
  While wrapping them with an asyncio loop requires little code, in some
  contexts it results in too much boilerplate.

  Testing async functions:

    class MyTest(unittest.TestCase):
      @run_in_asyncio_loop
      async def test_my_async_function(self):
        self.assertEqual(await async_function(), 42)

  Running the whole program in an event loop:

    @run_in_asyncio_loop
    async def main(argv):
      print('Hello world')

    if __name__ == '__main__':
      app.run(main)

  It is not advised to use this decorator beyond these two cases.

  Args:
    f: An async function to run in a loop.

  Returns:
    A synchronous function with the same arguments.
  """

  @functools.wraps(f)
  def decorated(*args, **kwargs) -> ReturnType:
    loop = asyncio.new_event_loop()
    asyncio.get_child_watcher().attach_loop(loop)
    return loop.run_until_complete(f(*args, **kwargs))

  return decorated


@functools.lru_cache()
def find_launch_script_path() -> str:
  """Finds the launch script path."""
  # We can get the launch script if it's provided explicitly, or when it's run
  # using a Python interpreter.
  launch_script_path = sys.argv[0]
  if hasattr(FLAGS, 'xm_launch_script') and FLAGS.xm_launch_script:
    launch_script_path = FLAGS.xm_launch_script
  if not launch_script_path.endswith('.py'):
    # If the launch script is built with subpar we are interested in the name
    # of the main module, rather than subpar binary.
    main_file_path = getattr(sys.modules['__main__'], '__file__', None)
    if (main_file_path and os.access(main_file_path, os.R_OK)):
      launch_script_path = main_file_path

  if not launch_script_path:
    return ''

  # The path may be relative, especially if it comes from sys.argv[0].
  return os.path.abspath(launch_script_path)


def resolve_path_relative_to_launcher(path: str) -> str:
  """Get the absolute assuming paths are relative to the launcher script file.

  Using this method a launcher script can refer to its own directory or parent
  directory via `.` and `..`.

  Args:
    path: Path that may be relative to the launch script.

  Returns:
    Absolute path.

  Raises:
    RuntimeError: If unable to determine the launch script path.
  """
  if os.path.isabs(path):
    return path

  launch_script_path = find_launch_script_path()
  if not os.access(launch_script_path, os.R_OK):
    raise RuntimeError(f'Unable to determine launch script path. '
                       f'The script is not present at {launch_script_path!r}. '
                       'This may happen if launch script changes the '
                       'working directory.')
  caller_file_path = os.path.realpath(launch_script_path)
  caller_dir = os.path.dirname(caller_file_path)
  return os.path.realpath(os.path.join(caller_dir, path))



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Definition of shared executable specifications."""

import abc
import os
import re
from typing import List, NamedTuple, Optional, Union

import attr
from xmanager.xm import job_blocks
from xmanager.xm import utils


def name_from_path(path: str) -> str:
  """Returns a safe to use executable name based on a filesystem path."""
  return re.sub('\\W', '_', os.path.basename(path.rstrip(os.sep)))


class ModuleName(NamedTuple):
  """Name of python module to execute when entering this project."""
  module_name: str


class CommandList(NamedTuple):
  """List of commands to execute when entering this project."""
  commands: List[str]


@attr.s(auto_attribs=True)
class Dockerfile(job_blocks.ExecutableSpec):
  """Dockerfile describes a Dockerfile for generating a docker image.

  This is a lower-level feature that could be solved using higher-level
  Executables such as BazelContainer or PythonContainer.

  Attributes:
    path: Specifies the build's context.
    dockerfile: The file that will be used for build instructions. Otherwise,
      {path}/Dockerfile will be used. Equivalent to `docker build -f`. A
      relative path will use a Dockerfile that is relative to the launcher
      script.
  """

  path: str = attr.ib(
      converter=utils.resolve_path_relative_to_launcher, default='.')
  dockerfile: str = attr.ib(
      # This field is always set once the object is initialized, so we use str
      # as type annotation. But the default value depends on another property
      # and is set in __attrs_post_init__, so we temporary convert None to ''.
      converter=lambda dockerfile: dockerfile or '',
      default=None)

  def __attrs_post_init__(self):
    if not self.dockerfile:
      self.dockerfile = os.path.join(self.path, 'Dockerfile')
    self.dockerfile = utils.resolve_path_relative_to_launcher(self.dockerfile)

  @property
  def name(self) -> str:
    return name_from_path(self.path)


@attr.s(auto_attribs=True)
class PythonContainer(job_blocks.ExecutableSpec):
  """PythonContainer describes a directory containing Python code.

  Attributes:
    entrypoint: The Python module or list of shell commands to run when entering
      this Python project.
    path: Relative or absolute path to the Python project. By default, the
      current directory (`'.'`) is used.
    base_image: Name of the image to initialize a new Docker build stage using
      the instruction `FROM`.
    docker_instructions: List of Docker instructions to apply when building the
      image. If not specified, the default one will be provided.

      When you use `docker_instructions`, you are responsible for copying the
      project directory. For example, if you are running with:

        path='/path/to/cifar10'

      You should include these steps in your `docker_instructions`:

        [
          'COPY cifar10/ cifar10',
          'WORKDIR cifar10',
        ]

      If your source code rarely changes, you can make this your first step.
      If you are frequently iterating on the source code, it is best practice
      to place these steps as late as possible in the list to maximize Docker
      layer-caching.
    use_deep_module: Whether the experiment code uses deep module structure
      (i.e., 'from <a.prefix> import models') or not (i.e., 'import models').

      If use_deep_module is set to True, and docker_instructions are used, it
      is recommended to use dedicated workdir and copy a whole project
      directory there. The example above should be modified as:

        [
          'RUN mkdir /workdir',
          'WORKDIR /workdir',
          'COPY cifar10/ /workdir/cifar10',
        ]
  """

  entrypoint: Union[ModuleName, CommandList]
  path: str = attr.ib(
      converter=utils.resolve_path_relative_to_launcher, default='.')
  base_image: Optional[str] = None
  docker_instructions: Optional[List[str]] = None
  use_deep_module: bool = False

  @property
  def name(self) -> str:
    return name_from_path(self.path)


class BinaryDependency(abc.ABC):
  """Additional resource for `Binary` / `BazelBinary`.

  Implementations can define backend-specific dependencies.
  """


@attr.s(auto_attribs=True)
class Container(job_blocks.ExecutableSpec):
  """A prebuilt Docker image.

  The image can be tagged locally or in a remote repository.

  Attributes:
    image_path: Path to a prebuilt container image.
  """

  image_path: str

  @property
  def name(self) -> str:
    return name_from_path(self.image_path)


@attr.s(auto_attribs=True)
class Binary(job_blocks.ExecutableSpec):
  """A prebuilt executable program.

  Attributes:
    path: Path to a prebuilt binary.
    dependencies: A list of data dependencies to be packaged together with the
      binary.
  """

  path: str
  dependencies: List[BinaryDependency] = attr.ib(
      converter=list, default=attr.Factory(list))

  @property
  def name(self) -> str:
    return name_from_path(self.path)


@attr.s(auto_attribs=True)
class BazelContainer(job_blocks.ExecutableSpec):
  """A Bazel target that produces a .tar image.

  Note that for targets based on https://github.com/bazelbuild/rules_docker one
  should append '.tar' to the label to specify a self-contained image.

  Attributes:
    label: The Bazel target to be built.
    bazel_args: Bazel command line arguments.
  """

  label: str
  bazel_args: List[str] = attr.ib(converter=list, default=attr.Factory(list))

  @property
  def name(self) -> str:
    return name_from_path(self.label)


@attr.s(auto_attribs=True)
class BazelBinary(job_blocks.ExecutableSpec):
  """A Bazel target that produces a self-contained binary.

  Note that for Python targets based on https://github.com/google/subpar
  a self-contained '.par' binary would be built.

  Attributes:
    label: The Bazel target to be built.
    dependencies: A list of data dependencies to be packaged together with
      the binary.
    bazel_args: Bazel command line arguments.
  """

  label: str
  dependencies: List[BinaryDependency] = attr.ib(
      converter=list, default=attr.Factory(list))
  bazel_args: List[str] = attr.ib(converter=list, default=attr.Factory(list))

  @property
  def name(self) -> str:
    return name_from_path(self.label)



# Copyright 2015 The Bazel Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyformat: disable
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# pylint: skip-file
# source: src/main/protobuf/invocation_policy.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='src/main/protobuf/invocation_policy.proto',
  package='blaze.invocation_policy',
  syntax='proto2',
  serialized_options=b'\n+com.google.devtools.build.lib.runtime.proto',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n)src/main/protobuf/invocation_policy.proto\x12\x17\x62laze.invocation_policy\"N\n\x10InvocationPolicy\x12:\n\rflag_policies\x18\x01 \x03(\x0b\x32#.blaze.invocation_policy.FlagPolicy\"\xb4\x02\n\nFlagPolicy\x12\x11\n\tflag_name\x18\x01 \x01(\t\x12\x10\n\x08\x63ommands\x18\x02 \x03(\t\x12\x36\n\tset_value\x18\x03 \x01(\x0b\x32!.blaze.invocation_policy.SetValueH\x00\x12:\n\x0buse_default\x18\x04 \x01(\x0b\x32#.blaze.invocation_policy.UseDefaultH\x00\x12\x42\n\x0f\x64isallow_values\x18\x05 \x01(\x0b\x32\'.blaze.invocation_policy.DisallowValuesH\x00\x12<\n\x0c\x61llow_values\x18\x06 \x01(\x0b\x32$.blaze.invocation_policy.AllowValuesH\x00\x42\x0b\n\toperation\"\xdf\x01\n\x08SetValue\x12\x12\n\nflag_value\x18\x01 \x03(\t\x12\x13\n\x0boverridable\x18\x02 \x01(\x08\x12\x0e\n\x06\x61ppend\x18\x03 \x01(\x08\x12<\n\x08\x62\x65havior\x18\x04 \x01(\x0e\x32*.blaze.invocation_policy.SetValue.Behavior\"\\\n\x08\x42\x65havior\x12\r\n\tUNDEFINED\x10\x00\x12\x13\n\x0f\x41LLOW_OVERRIDES\x10\x01\x12\n\n\x06\x41PPEND\x10\x02\x12 \n\x1c\x46INAL_VALUE_IGNORE_OVERRIDES\x10\x03\"\x0c\n\nUseDefault\"\x97\x01\n\x0e\x44isallowValues\x12\x19\n\x11\x64isallowed_values\x18\x01 \x03(\t\x12\x13\n\tnew_value\x18\x03 \x01(\tH\x00\x12:\n\x0buse_default\x18\x04 \x01(\x0b\x32#.blaze.invocation_policy.UseDefaultH\x00\x42\x13\n\x11replacement_valueJ\x04\x08\x02\x10\x03\"\x91\x01\n\x0b\x41llowValues\x12\x16\n\x0e\x61llowed_values\x18\x01 \x03(\t\x12\x13\n\tnew_value\x18\x03 \x01(\tH\x00\x12:\n\x0buse_default\x18\x04 \x01(\x0b\x32#.blaze.invocation_policy.UseDefaultH\x00\x42\x13\n\x11replacement_valueJ\x04\x08\x02\x10\x03\x42-\n+com.google.devtools.build.lib.runtime.proto'
)



_SETVALUE_BEHAVIOR = _descriptor.EnumDescriptor(
  name='Behavior',
  full_name='blaze.invocation_policy.SetValue.Behavior',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNDEFINED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ALLOW_OVERRIDES', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='APPEND', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FINAL_VALUE_IGNORE_OVERRIDES', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=593,
  serialized_end=685,
)
_sym_db.RegisterEnumDescriptor(_SETVALUE_BEHAVIOR)


_INVOCATIONPOLICY = _descriptor.Descriptor(
  name='InvocationPolicy',
  full_name='blaze.invocation_policy.InvocationPolicy',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='flag_policies', full_name='blaze.invocation_policy.InvocationPolicy.flag_policies', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=70,
  serialized_end=148,
)


_FLAGPOLICY = _descriptor.Descriptor(
  name='FlagPolicy',
  full_name='blaze.invocation_policy.FlagPolicy',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='flag_name', full_name='blaze.invocation_policy.FlagPolicy.flag_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='commands', full_name='blaze.invocation_policy.FlagPolicy.commands', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='set_value', full_name='blaze.invocation_policy.FlagPolicy.set_value', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='use_default', full_name='blaze.invocation_policy.FlagPolicy.use_default', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='disallow_values', full_name='blaze.invocation_policy.FlagPolicy.disallow_values', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='allow_values', full_name='blaze.invocation_policy.FlagPolicy.allow_values', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='operation', full_name='blaze.invocation_policy.FlagPolicy.operation',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=151,
  serialized_end=459,
)


_SETVALUE = _descriptor.Descriptor(
  name='SetValue',
  full_name='blaze.invocation_policy.SetValue',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='flag_value', full_name='blaze.invocation_policy.SetValue.flag_value', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='overridable', full_name='blaze.invocation_policy.SetValue.overridable', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='append', full_name='blaze.invocation_policy.SetValue.append', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='behavior', full_name='blaze.invocation_policy.SetValue.behavior', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SETVALUE_BEHAVIOR,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=462,
  serialized_end=685,
)


_USEDEFAULT = _descriptor.Descriptor(
  name='UseDefault',
  full_name='blaze.invocation_policy.UseDefault',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=687,
  serialized_end=699,
)


_DISALLOWVALUES = _descriptor.Descriptor(
  name='DisallowValues',
  full_name='blaze.invocation_policy.DisallowValues',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='disallowed_values', full_name='blaze.invocation_policy.DisallowValues.disallowed_values', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='new_value', full_name='blaze.invocation_policy.DisallowValues.new_value', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='use_default', full_name='blaze.invocation_policy.DisallowValues.use_default', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='replacement_value', full_name='blaze.invocation_policy.DisallowValues.replacement_value',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=702,
  serialized_end=853,
)


_ALLOWVALUES = _descriptor.Descriptor(
  name='AllowValues',
  full_name='blaze.invocation_policy.AllowValues',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='allowed_values', full_name='blaze.invocation_policy.AllowValues.allowed_values', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='new_value', full_name='blaze.invocation_policy.AllowValues.new_value', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='use_default', full_name='blaze.invocation_policy.AllowValues.use_default', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='replacement_value', full_name='blaze.invocation_policy.AllowValues.replacement_value',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=856,
  serialized_end=1001,
)

_INVOCATIONPOLICY.fields_by_name['flag_policies'].message_type = _FLAGPOLICY
_FLAGPOLICY.fields_by_name['set_value'].message_type = _SETVALUE
_FLAGPOLICY.fields_by_name['use_default'].message_type = _USEDEFAULT
_FLAGPOLICY.fields_by_name['disallow_values'].message_type = _DISALLOWVALUES
_FLAGPOLICY.fields_by_name['allow_values'].message_type = _ALLOWVALUES
_FLAGPOLICY.oneofs_by_name['operation'].fields.append(
  _FLAGPOLICY.fields_by_name['set_value'])
_FLAGPOLICY.fields_by_name['set_value'].containing_oneof = _FLAGPOLICY.oneofs_by_name['operation']
_FLAGPOLICY.oneofs_by_name['operation'].fields.append(
  _FLAGPOLICY.fields_by_name['use_default'])
_FLAGPOLICY.fields_by_name['use_default'].containing_oneof = _FLAGPOLICY.oneofs_by_name['operation']
_FLAGPOLICY.oneofs_by_name['operation'].fields.append(
  _FLAGPOLICY.fields_by_name['disallow_values'])
_FLAGPOLICY.fields_by_name['disallow_values'].containing_oneof = _FLAGPOLICY.oneofs_by_name['operation']
_FLAGPOLICY.oneofs_by_name['operation'].fields.append(
  _FLAGPOLICY.fields_by_name['allow_values'])
_FLAGPOLICY.fields_by_name['allow_values'].containing_oneof = _FLAGPOLICY.oneofs_by_name['operation']
_SETVALUE.fields_by_name['behavior'].enum_type = _SETVALUE_BEHAVIOR
_SETVALUE_BEHAVIOR.containing_type = _SETVALUE
_DISALLOWVALUES.fields_by_name['use_default'].message_type = _USEDEFAULT
_DISALLOWVALUES.oneofs_by_name['replacement_value'].fields.append(
  _DISALLOWVALUES.fields_by_name['new_value'])
_DISALLOWVALUES.fields_by_name['new_value'].containing_oneof = _DISALLOWVALUES.oneofs_by_name['replacement_value']
_DISALLOWVALUES.oneofs_by_name['replacement_value'].fields.append(
  _DISALLOWVALUES.fields_by_name['use_default'])
_DISALLOWVALUES.fields_by_name['use_default'].containing_oneof = _DISALLOWVALUES.oneofs_by_name['replacement_value']
_ALLOWVALUES.fields_by_name['use_default'].message_type = _USEDEFAULT
_ALLOWVALUES.oneofs_by_name['replacement_value'].fields.append(
  _ALLOWVALUES.fields_by_name['new_value'])
_ALLOWVALUES.fields_by_name['new_value'].containing_oneof = _ALLOWVALUES.oneofs_by_name['replacement_value']
_ALLOWVALUES.oneofs_by_name['replacement_value'].fields.append(
  _ALLOWVALUES.fields_by_name['use_default'])
_ALLOWVALUES.fields_by_name['use_default'].containing_oneof = _ALLOWVALUES.oneofs_by_name['replacement_value']
DESCRIPTOR.message_types_by_name['InvocationPolicy'] = _INVOCATIONPOLICY
DESCRIPTOR.message_types_by_name['FlagPolicy'] = _FLAGPOLICY
DESCRIPTOR.message_types_by_name['SetValue'] = _SETVALUE
DESCRIPTOR.message_types_by_name['UseDefault'] = _USEDEFAULT
DESCRIPTOR.message_types_by_name['DisallowValues'] = _DISALLOWVALUES
DESCRIPTOR.message_types_by_name['AllowValues'] = _ALLOWVALUES
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

InvocationPolicy = _reflection.GeneratedProtocolMessageType('InvocationPolicy', (_message.Message,), {
  'DESCRIPTOR' : _INVOCATIONPOLICY,
  '__module__' : 'src.main.protobuf.invocation_policy_pb2'
  # @@protoc_insertion_point(class_scope:blaze.invocation_policy.InvocationPolicy)
  })
_sym_db.RegisterMessage(InvocationPolicy)

FlagPolicy = _reflection.GeneratedProtocolMessageType('FlagPolicy', (_message.Message,), {
  'DESCRIPTOR' : _FLAGPOLICY,
  '__module__' : 'src.main.protobuf.invocation_policy_pb2'
  # @@protoc_insertion_point(class_scope:blaze.invocation_policy.FlagPolicy)
  })
_sym_db.RegisterMessage(FlagPolicy)

SetValue = _reflection.GeneratedProtocolMessageType('SetValue', (_message.Message,), {
  'DESCRIPTOR' : _SETVALUE,
  '__module__' : 'src.main.protobuf.invocation_policy_pb2'
  # @@protoc_insertion_point(class_scope:blaze.invocation_policy.SetValue)
  })
_sym_db.RegisterMessage(SetValue)

UseDefault = _reflection.GeneratedProtocolMessageType('UseDefault', (_message.Message,), {
  'DESCRIPTOR' : _USEDEFAULT,
  '__module__' : 'src.main.protobuf.invocation_policy_pb2'
  # @@protoc_insertion_point(class_scope:blaze.invocation_policy.UseDefault)
  })
_sym_db.RegisterMessage(UseDefault)

DisallowValues = _reflection.GeneratedProtocolMessageType('DisallowValues', (_message.Message,), {
  'DESCRIPTOR' : _DISALLOWVALUES,
  '__module__' : 'src.main.protobuf.invocation_policy_pb2'
  # @@protoc_insertion_point(class_scope:blaze.invocation_policy.DisallowValues)
  })
_sym_db.RegisterMessage(DisallowValues)

AllowValues = _reflection.GeneratedProtocolMessageType('AllowValues', (_message.Message,), {
  'DESCRIPTOR' : _ALLOWVALUES,
  '__module__' : 'src.main.protobuf.invocation_policy_pb2'
  # @@protoc_insertion_point(class_scope:blaze.invocation_policy.AllowValues)
  })
_sym_db.RegisterMessage(AllowValues)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for cloud_build."""
from absl.testing import absltest

from xmanager.cloud import cloud_build


class CloudBuildTest(absltest.TestCase):

  def test_build_request_body(self):
    client = cloud_build.Client(
        'my-project',
        'my-bucket',
        'fake-creds',
        use_kaniko=False,
        use_cloud_build_cache=False)
    image = client._build_request_body('path/to/project', 'my-image', 'live')
    self.assertEqual(
        image, {
            'images': ['my-image'],
            'options': {
                'machineType': 'E2_HIGHCPU_32'
            },
            'source': {
                'storageSource': {
                    'bucket': 'my-bucket',
                    'object': 'path/to/project',
                },
            },
            'steps': [{
                'args': [
                    'build', '-t', 'my-image:live', '-t', 'my-image:latest', '.'
                ],
                'name': 'gcr.io/cloud-builders/docker',
            }],
            'timeout': '1200s'
        })

  def test_build_request_body_use_kaniko(self):
    client = cloud_build.Client(
        'my-project',
        'my-bucket',
        'fake-creds',
        use_kaniko=True,
        use_cloud_build_cache=False)
    image = client._build_request_body('path/to/project', 'my-image', 'live')
    self.assertEqual(
        image, {
            'source': {
                'storageSource': {
                    'bucket': 'my-bucket',
                    'object': 'path/to/project',
                },
            },
            'steps': [{
                'args': [
                    '--destination=my-image:live',
                    '--destination=my-image:latest',
                    '--cache=true',
                    '--cache-ttl=336h',
                ],
                'name': 'gcr.io/kaniko-project/executor:latest',
            }],
            'timeout': '1200s'
        })

  def test_build_request_body_use_build_cache(self):
    client = cloud_build.Client(
        'my-project',
        'my-bucket',
        'fake-creds',
        use_kaniko=False,
        use_cloud_build_cache=True)
    image = client._build_request_body('path/to/project', 'my-image', 'live')
    self.assertEqual(
        image, {
            'images': ['my-image'],
            'options': {
                'machineType': 'E2_HIGHCPU_32'
            },
            'source': {
                'storageSource': {
                    'bucket': 'my-bucket',
                    'object': 'path/to/project',
                },
            },
            'steps': [{
                'args': [
                    'build',
                    '-t',
                    'my-image:live',
                    '-t',
                    'my-image:latest',
                    '--cache-from',
                    'my-image:latest',
                    '.',
                ],
                'name': 'gcr.io/cloud-builders/docker',
            }],
            'timeout': '1200s'
        })


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementation of the local scheduler experiment."""

import asyncio
from concurrent import futures
import time
from typing import Any, Awaitable, Callable, List, Mapping, Sequence

from absl import logging
import attr
from kubernetes import client as k8s_client
from xmanager import xm
from xmanager.cloud import caip
from xmanager.cloud import kubernetes
from xmanager.xm import async_packager
from xmanager.xm import id_predictor
from xmanager.xm import job_operators
from xmanager.xm import pattern_matching
from xmanager.xm_local import execution as local_execution
from xmanager.xm_local import executors as local_executors
from xmanager.xm_local import status as local_status
from xmanager.xm_local.packaging import router as packaging_router
from xmanager.xm_local.storage import database


def _throw_on_unknown_executor(job: xm.Job, executor: Any):
  raise TypeError(f'Unsupported executor: {executor!r}. Job: {job!r}')


_EXECUTOR_VALIDATOR = pattern_matching.match(
    pattern_matching.Case([xm.Job, local_executors.Local], lambda *_: None),
    pattern_matching.Case([xm.Job, local_executors.Caip], lambda *_: None),
    pattern_matching.Case([xm.Job, local_executors.Kubernetes],
                          lambda *_: None),
    _throw_on_unknown_executor,
)


def _validate_job_group(job_group: xm.JobGroup) -> None:
  all_jobs = job_operators.flatten_jobs(job_group)
  for job in all_jobs:
    _EXECUTOR_VALIDATOR(job, job.executor)


@attr.s(auto_attribs=True)
class _LaunchResult:
  caip_handles: List[caip.CaipHandle]
  k8s_handles: List[kubernetes.KubernetesHandle]
  local_handles: List[local_execution.LocalExecutionHandle]


class LocalExperimentUnit(xm.ExperimentUnit):
  """Experiment unit operated by the local backend."""

  def __init__(self, experiment: 'LocalExperiment', experiment_title: str,
               create_task: Callable[[Awaitable[Any]], futures.Future],
               args: Mapping[str, Any], role: xm.ExperimentUnitRole) -> None:
    super().__init__(experiment, create_task, args, role)
    self._experiment_title = experiment_title
    self._local_execution_handles: List[
        local_execution.LocalExecutionHandle] = []
    self._non_local_execution_handles: List[
        local_execution.ExecutionHandle] = []

  async def _submit_jobs_for_execution(self,
                                       job_group: xm.JobGroup) -> _LaunchResult:
    # We are delegating the traversal of the job group to modules.
    # That improves modularity, but sacrifices the ability to make
    # cross-executor decisions.
    caip_handles = caip.launch(self._experiment_title,
                               self.experiment_unit_name, job_group)
    k8s_handles = kubernetes.launch(self.get_full_job_name, job_group)
    local_handles = await local_execution.launch(self.get_full_job_name,
                                                 job_group)
    return _LaunchResult(
        caip_handles=caip_handles,
        k8s_handles=k8s_handles,
        local_handles=local_handles,
    )

  def _ingest_execution_handles(self, launch_result: _LaunchResult) -> None:
    self._non_local_execution_handles.extend(launch_result.caip_handles +
                                             launch_result.k8s_handles)
    self._local_execution_handles.extend(launch_result.local_handles)

  def _monitor_local_jobs(
      self,
      local_execution_handles: Sequence[local_execution.LocalExecutionHandle]
  ) -> None:
    for handle in local_execution_handles:
      self._create_task(handle.monitor())

  async def _wait_until_complete(self) -> None:
    try:
      await asyncio.gather(*[
          handle.wait() for handle in self._local_execution_handles +
          self._non_local_execution_handles
      ])
    except RuntimeError as error:
      raise xm.ExperimentUnitFailedError(error)

  async def wait_for_local_jobs(self, is_exit_abrupt: bool):
    if not is_exit_abrupt:
      await asyncio.gather(
          *[handle.wait() for handle in self._local_execution_handles])

  def stop(self) -> None:
    """Initiate the process to stop the work unit from running.

    This method will synchronously make a request for the work unit to stop.
    However, the method does not actually wait for the work unit to be in a
    terminal state.

    Use self.wait_until_complete() after self.stop() to guarantee the work unit
    is stopped.
    """

    def stop_caip_handle(caip_handle: caip.CaipHandle) -> None:
      caip_handle.stop()

    def throw_on_unknown_handle(handle: Any) -> None:
      raise TypeError(f'Unsupported handle: {handle!r}')

    handle_stopper = pattern_matching.match(
        stop_caip_handle,
        throw_on_unknown_handle,
    )
    handles = self._non_local_execution_handles + self._local_execution_handles
    for handle in handles:
      handle_stopper(handle)

  def get_status(self) -> local_status.LocalWorkUnitStatus:
    """Gets the current status of the work unit."""
    handles = self._non_local_execution_handles + self._local_execution_handles
    if len(handles) == 1:
      return handles[0].get_status()
    raise NotImplementedError(
        'Status aggregation for work units with multiple jobs is not '
        'implemented yet.')


class LocalWorkUnit(LocalExperimentUnit):
  """A work unit operated by the local backend."""

  def __init__(self, experiment: 'LocalExperiment', experiment_title: str,
               create_task: Callable[[Awaitable[Any]], futures.Future],
               args: Mapping[str, Any], role: xm.ExperimentUnitRole,
               work_unit_id_predictor: id_predictor.Predictor) -> None:
    super().__init__(experiment, experiment_title, create_task, args, role)
    self._work_unit_id_predictor = work_unit_id_predictor
    self._work_unit_id = self._work_unit_id_predictor.reserve_id()

  def _save_handles_to_storage(
      self, handles: Sequence[local_execution.ExecutionHandle]) -> None:
    """Saves jobs present in the handlers."""

    def save_caip_handle(caip_handle: caip.CaipHandle) -> None:
      database.database().insert_caip_job(self.experiment_id, self.work_unit_id,
                                          self.experiment_unit_name,
                                          caip_handle.job_name)

    def save_k8s_handle(k8s_handle: kubernetes.KubernetesHandle) -> None:
      for job in k8s_handle.jobs:
        namespace = job.metadata.namespace or 'default'
        name = job.metadata.name
        database.database().insert_kubernetes_job(self.experiment_id,
                                                  self.work_unit_id,
                                                  self.experiment_unit_name,
                                                  namespace, name)

    def throw_on_unknown_handle(handle: Any) -> None:
      raise TypeError(f'Unsupported handle: {handle!r}')

    handle_saver = pattern_matching.match(
        save_caip_handle,
        save_k8s_handle,
        throw_on_unknown_handle,
    )
    for handle in handles:
      handle_saver(handle)

  async def _launch_job_group(self, job_group: xm.JobGroup,
                              args_view: Mapping[str, Any]) -> None:
    del args_view  # Unused.
    _validate_job_group(job_group)

    async with self._work_unit_id_predictor.submit_id(self.work_unit_id):
      launch_result = await self._submit_jobs_for_execution(job_group)
      self._ingest_execution_handles(launch_result)
      # TODO: Save the local jobs to the database as well.
      self._save_handles_to_storage(launch_result.caip_handles +
                                    launch_result.k8s_handles)
      self._monitor_local_jobs(launch_result.local_handles)

  @property
  def experiment_unit_name(self) -> str:
    return f'{self.experiment_id}_{self._work_unit_id}'

  @property
  def work_unit_id(self) -> int:
    return self._work_unit_id


class LocalAuxiliaryUnit(LocalExperimentUnit):
  """An auxiliary unit operated by the local backend."""

  async def _launch_job_group(self, job_group: xm.JobGroup,
                              args_view: Mapping[str, Any]) -> None:
    del args_view  # Unused.
    _validate_job_group(job_group)

    launch_result = await self._submit_jobs_for_execution(job_group)
    self._ingest_execution_handles(launch_result)
    self._monitor_local_jobs(launch_result.local_handles)

  @property
  def experiment_unit_name(self) -> str:
    return f'{self.experiment_id}_auxiliary'


class LocalExperiment(xm.Experiment):
  """Experiment contains a family of jobs that run with the local scheduler."""

  _id: int
  _experiment_title: str
  _experiment_units: List[LocalExperimentUnit]
  _work_unit_count: int
  _async_packager = async_packager.AsyncPackager(packaging_router.package)

  def __init__(self, experiment_title: str) -> None:
    super().__init__()
    # To distinguish local job names until we use a local database generator.
    self._id = int(time.time() * 10**3)
    self._experiment_title = experiment_title
    self._experiment_units = []
    self._work_unit_count = 0

  def _create_experiment_unit(self, args: Mapping[str, Any],
                              role: xm.ExperimentUnitRole) -> xm.ExperimentUnit:
    """Creates a new WorkUnit instance for the experiment."""

    def create_work_unit(role: xm.WorkUnitRole):
      work_unit = LocalWorkUnit(
          self,
          self._experiment_title,
          self._create_task,
          args,
          role,
          self._work_unit_id_predictor,
      )
      self._experiment_units.append(work_unit)
      self._work_unit_count += 1
      database.database().insert_work_unit(
          self.experiment_id,
          work_unit.work_unit_id,
      )
      return work_unit

    # TODO: Support `role.termination_delay_secs`.
    def create_auxiliary_unit(role: xm.AuxiliaryUnitRole):
      auxiliary_unit = LocalAuxiliaryUnit(
          self,
          self._experiment_title,
          self._create_task,
          args,
          role,
      )
      self._experiment_units.append(auxiliary_unit)
      return auxiliary_unit

    return pattern_matching.match(create_work_unit, create_auxiliary_unit)(role)

  def _wait_for_local_jobs(self, is_exit_abrupt: bool):
    if self._experiment_units:
      print('Waiting for local jobs to complete. '
            'Press Ctrl+C to terminate them and exit')
    for unit in self._experiment_units:
      self._create_task(unit.wait_for_local_jobs(is_exit_abrupt))

  def __exit__(self, exc_type, exc_value, traceback):
    # Flush `.add` calls.
    self._wait_for_tasks()
    self._wait_for_local_jobs(exc_value is not None)
    return super().__exit__(exc_type, exc_value, traceback)

  async def __aexit__(self, exc_type, exc_value, traceback):
    # Flush `.add` calls.
    await self._await_for_tasks()
    self._wait_for_local_jobs(exc_value is not None)
    return await super().__aexit__(exc_type, exc_value, traceback)

  @property
  def experiment_id(self) -> int:
    return self._id

  @property
  def work_unit_count(self) -> int:
    return self._work_unit_count

  @property
  def work_units(self) -> Mapping[int, LocalExperimentUnit]:
    """Gets work units created via self.add()."""
    raise NotImplementedError


def create_experiment(experiment_title: str) -> xm.Experiment:
  """Create Experiment."""
  experiment = LocalExperiment(experiment_title)
  database.database().insert_experiment(experiment.experiment_id,
                                        experiment._experiment_title)  # pylint: disable=protected-access
  return experiment


def get_experiment(experiment_id: int) -> xm.Experiment:
  """Returns an Experiment instance associated with this experiment id."""
  # pylint: disable=protected-access
  experiment_result = database.database().get_experiment(experiment_id)
  experiment = LocalExperiment(experiment_result.experiment_title)
  experiment._id = experiment_id
  experiment._work_unit_id_predictor = id_predictor.Predictor(1)
  for work_unit_result in experiment_result.work_units:
    work_unit = LocalWorkUnit(experiment, experiment_result.experiment_title,
                              lambda _: None, {}, xm.WorkUnitRole(),
                              experiment._work_unit_id_predictor)
    work_unit._work_unit_id = work_unit_result.work_unit_id
    non_local_handles = []
    kubernetes_jobs = []
    for _, data in work_unit_result.jobs.items():
      if data.HasField('local'):
        logging.warning(
            '[Experiment id: %s, work unit id: %s] '
            'Loading local experiment units from storage is not implemented.',
            experiment_id, work_unit_result.work_unit_id)
      elif data.HasField('caip'):
        non_local_handles = [caip.CaipHandle(data.caip.resource_name)]
      elif data.HasField('kubernetes'):
        job = k8s_client.V1Job()
        job.metadata = k8s_client.V1ObjectMeta(
            namespace=data.kubernetes.namespace, name=data.kubernetes.job_name)
        kubernetes_jobs.append(job)
        non_local_handles = [kubernetes.KubernetesHandle(kubernetes_jobs)]
    work_unit._non_local_execution_handles = non_local_handles
    experiment._experiment_units.append(work_unit)
    experiment._work_unit_count += 1
  return experiment
  # pylint: enable=protected-access


def list_experiments() -> Sequence[xm.Experiment]:
  """Yields a list of Experiment instances that have been created thus far."""
  experiment_ids = database.database().list_experiment_ids()
  return [get_experiment(experiment_id) for experiment_id in experiment_ids]



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Database connector module."""
import abc
import functools
import os
from typing import Dict, List

import attr
import sqlalchemy
import sqlite3
from xmanager.generated import data_pb2

from google.protobuf import text_format


@attr.s(auto_attribs=True)
class WorkUnitResult:
  """Result of a WorkUnit database query."""
  work_unit_id: int
  jobs: Dict[str, data_pb2.Job]


@attr.s(auto_attribs=True)
class ExperimentResult:
  """Result of an Experiment database query."""
  experiment_id: int
  experiment_title: str
  work_units: List[WorkUnitResult]


Engine = sqlalchemy.engine.Engine


class SqlSettings(abc.ABC):
  """Settings for a SQL dialect."""

  @abc.abstractmethod
  def create_engine(self) -> Engine:
    raise NotImplementedError

  @abc.abstractmethod
  def execute_script(self, script: str) -> None:
    raise NotImplementedError


@attr.s(auto_attribs=True)
class SqliteSettings(SqlSettings):
  """Settings for the Sqlite dialect."""

  path: str = os.path.expanduser('~/.xmanager/experiments.sqlite3')

  def create_engine(self) -> Engine:
    if not os.path.isdir(os.path.dirname(self.path)):
      os.makedirs(os.path.dirname(self.path))
    if not os.path.isfile(self.path):
      sqlite3.connect(self.path)
    return sqlalchemy.create_engine(f'sqlite:///{self.path}')

  def execute_script(self, script: str) -> None:
    with open(script) as f:
      content = f.read()
    conn = sqlite3.connect(self.path)
    cursor = conn.cursor()
    cursor.executescript(content)


@functools.lru_cache()
def database():
  # Create only a single global singleton for database access.
  return Database()


class Database:
  """Database object with interacting with experiment metadata storage."""

  def __init__(self, settings: SqlSettings = SqliteSettings()):
    self.settings = settings
    self.engine: Engine = settings.create_engine()
    self.maybe_migrate_database_version(self.engine)

  def maybe_migrate_database_version(self, engine: Engine):
    """Check the database VersionHistory table and maybe migrate."""
    # Create the tables for the first time.
    if 'VersionHistory' not in engine.table_names():
      schema = os.path.join(
          os.path.dirname(os.path.realpath(__file__)), 'schema.sql')
      self.settings.execute_script(schema)

    rows = list(
        self.engine.execute(
            'SELECT Version, Timestamp FROM VersionHistory ORDER BY Timestamp DESC LIMIT 1'
        ))
    if not rows:
      raise ValueError('The database is invalid. It has no VersionHistory.')
    if rows[0][0] > 1:
      raise ValueError(
          f'The database schema is on an unsupported version: {rows[0][0]}')

  def insert_experiment(self, experiment_id: int,
                        experiment_title: str) -> None:
    query = ('INSERT INTO Experiment (Id, Title) '
             'VALUES (:experiment_id, :experiment_title)')
    self.engine.execute(
        query, experiment_id=experiment_id, experiment_title=experiment_title)

  def insert_work_unit(self, experiment_id: int, work_unit_id: int) -> None:
    query = ('INSERT INTO WorkUnit (ExperimentId, WorkUnitId) '
             'VALUES (:experiment_id, :work_unit_id)')
    self.engine.execute(
        query, experiment_id=experiment_id, work_unit_id=work_unit_id)

  def insert_caip_job(self, experiment_id: int, work_unit_id: int, name: str,
                      caip_job_id: str) -> None:
    job = data_pb2.Job(caip=data_pb2.AIPlatformJob(resource_name=caip_job_id))
    data = text_format.MessageToBytes(job)
    query = ('INSERT INTO Job (ExperimentId, WorkUnitId, Name, Data) '
             'VALUES (:experiment_id, :work_unit_id, :name, :data)')
    self.engine.execute(
        query,
        experiment_id=experiment_id,
        work_unit_id=work_unit_id,
        name=name,
        data=data)

  def insert_kubernetes_job(self, experiment_id: int, work_unit_id: int,
                            name: str, namespace: str, job_name: str) -> None:
    job = data_pb2.Job(
        kubernetes=data_pb2.KubernetesJob(
            namespace=namespace, job_name=job_name))
    data = text_format.MessageToString(job)
    query = ('INSERT INTO Job (ExperimentId, WorkUnitId, Name, Data) '
             'VALUES (:experiment_id, :work_unit_id, :name, :data)')
    self.engine.execute(
        query,
        experiment_id=experiment_id,
        work_unit_id=work_unit_id,
        name=name,
        data=data)

  def list_experiment_ids(self) -> List[int]:
    """Lists all the experiment ids from local database."""
    query = ('SELECT Id FROM Experiment')
    rows = self.engine.execute(query)
    return [r['Id'] for r in rows]

  def get_experiment(self, experiment_id: int) -> ExperimentResult:
    """Gets an experiment from local database."""
    query = ('SELECT Title FROM Experiment WHERE Id=:experiment_id')
    rows = self.engine.execute(query, experiment_id=experiment_id)
    title = None
    for r in rows:
      title = r['Title']
      break
    if title is None:
      raise ValueError(f'Experiment Id {experiment_id} doesn\'t exist.')
    return ExperimentResult(experiment_id, title,
                            self.list_work_units(experiment_id))

  def list_work_units(self, experiment_id: int) -> List[WorkUnitResult]:
    """Lists an experiment's work unit ids from local database."""
    query = ('SELECT WorkUnitId '
             'FROM WorkUnit WHERE ExperimentId=:experiment_id')
    rows = self.engine.execute(query, experiment_id=experiment_id)
    return [self.get_work_unit(experiment_id, r['WorkUnitId']) for r in rows]

  def get_work_unit(self, experiment_id: int,
                    work_unit_id: int) -> WorkUnitResult:
    """Gets a work unit from local database."""
    query = ('SELECT Name, Data FROM Job '
             'WHERE ExperimentId=:experiment_id AND WorkUnitID=:work_unit_id')
    rows = self.engine.execute(
        query, experiment_id=experiment_id, work_unit_id=work_unit_id)
    jobs = {}
    for r in rows:
      job = data_pb2.Job()
      jobs[r['Name']] = text_format.Parse(r['Data'], job)
    return WorkUnitResult(work_unit_id, jobs)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Client for interacting with AI Platform (Unified).

https://cloud.google.com/ai-platform-unified/docs/reference/rest
"""
import asyncio
import functools
import logging
import math
import os
import time
from typing import Any, Dict, List, Optional, Sequence, Tuple

import attr
from google.cloud import aiplatform
from google.cloud import aiplatform_v1 as aip_v1
from google.cloud import aiplatform_v1beta1 as aip_v1beta
from xmanager import xm
from xmanager.cloud import auth
from xmanager.xm import utils
from xmanager.xm_local import executables as local_executables
from xmanager.xm_local import execution as local_execution
from xmanager.xm_local import executors as local_executors
from xmanager.xm_local import status as local_status

_DEFAULT_LOCATION = 'us-central1'
# The only machines available on AI Platform are N1 machines.
# https://cloud.google.com/ai-platform-unified/docs/predictions/machine-types#machine_type_comparison
# TODO: Add machines that support A100.
# TODO: Move lookup to a canonical place.
_MACHINE_TYPE_TO_CPU_RAM = {
    'n1-standard-4': (4, 15 * xm.GiB),
    'n1-standard-8': (8, 30 * xm.GiB),
    'n1-standard-16': (16, 60 * xm.GiB),
    'n1-standard-32': (32, 120 * xm.GiB),
    'n1-standard-64': (64, 240 * xm.GiB),
    'n1-standard-96': (96, 360 * xm.GiB),
    'n1-highmem-2': (2, 13 * xm.GiB),
    'n1-highmem-4': (4, 26 * xm.GiB),
    'n1-highmem-8': (8, 52 * xm.GiB),
    'n1-highmem-16': (16, 104 * xm.GiB),
    'n1-highmem-32': (32, 208 * xm.GiB),
    'n1-highmem-64': (64, 416 * xm.GiB),
    'n1-highmem-96': (96, 624 * xm.GiB),
    'n1-highcpu-16': (16, 14 * xm.GiB),
    'n1-highcpu-32': (32, 28 * xm.GiB),
    'n1-highcpu-64': (64, 57 * xm.GiB),
    'n1-highcpu-96': (96, 86 * xm.GiB),
}

_CLOUD_TPU_ACCELERATOR_TYPES = {
    xm.ResourceType.TPU_V2: 'TPU_V2',
    xm.ResourceType.TPU_V3: 'TPU_V3',
}

_STATE_TO_STATUS = {
    aip_v1.JobState.JOB_STATE_SUCCEEDED:
        local_status.LocalWorkUnitStatusEnum.COMPLETED,
    aip_v1.JobState.JOB_STATE_CANCELLED:
        local_status.LocalWorkUnitStatusEnum.CANCELLED,
    aip_v1.JobState.JOB_STATE_QUEUED:
        local_status.LocalWorkUnitStatusEnum.RUNNING,
    aip_v1.JobState.JOB_STATE_PENDING:
        local_status.LocalWorkUnitStatusEnum.RUNNING,
    aip_v1.JobState.JOB_STATE_RUNNING:
        local_status.LocalWorkUnitStatusEnum.RUNNING,
    aip_v1.JobState.JOB_STATE_CANCELLING:
        local_status.LocalWorkUnitStatusEnum.RUNNING,
    aip_v1.JobState.JOB_STATE_PAUSED:
        local_status.LocalWorkUnitStatusEnum.RUNNING,
    aip_v1.JobState.JOB_STATE_FAILED:
        local_status.LocalWorkUnitStatusEnum.FAILED,
}

# Hide noisy warning regarding:
# `file_cache is unavailable when using oauth2client >= 4.0.0 or google-auth`
logging.getLogger('googleapiclient.discovery_cache').setLevel(logging.ERROR)


@functools.lru_cache()
def client():
  # Global singleton defers client creation until an actual launch.
  # If the user only launches local jobs, they don't need cloud access.
  return Client()


def _caip_job_predicate(job: xm.Job) -> bool:
  return isinstance(job.executor, local_executors.Caip)


class Client:
  """Client class for interacting with AI Platform (Unified)."""

  def __init__(self,
               project: Optional[str] = None,
               location: str = _DEFAULT_LOCATION) -> None:
    """Create a client.

    Args:
      project: GCP project ID.
      location: Location of the API endpoint. Defaults to region us-central1.
        https://cloud.google.com/vertex-ai/docs/reference/rest#service-endpoint
    """
    self.location = location
    self.project = project or auth.get_project_name()
    self.parent = f'projects/{self.project}/locations/{self.location}'
    # TODO: move staging_bucket when issue is fixed
    # https://github.com/googleapis/python-aiplatform/pull/421
    aiplatform.init(
        project=self.project,
        location=self.location,
        staging_bucket=f'gs://{auth.get_bucket()}')

  def launch(self, name: str, jobs: Sequence[xm.Job]) -> str:
    """Launch jobs on AI Platform (Unified)."""
    pools = []
    tensorboard, output_dir = self.get_tensorboard_settings(jobs)
    for i, job in enumerate(jobs):
      executable = job.executable
      if not isinstance(executable,
                        local_executables.GoogleContainerRegistryImage):
        raise ValueError('Executable {} has type {}. Executable must be of '
                         'type GoogleContainerRegistryImage'.format(
                             executable, type(executable)))

      args = xm.merge_args(executable.args, job.args).to_list(utils.ARG_ESCAPER)
      env_vars = {**executable.env_vars, **job.env_vars}
      env = [{'name': k, 'value': v} for k, v in env_vars.items()]
      if job.executor.requirements.accelerator in xm.TpuType:
        tpu_runtime_version = 'nightly'  # pylint: disable=unused-variable
        if job.executor.tpu_capability:
          tpu_runtime_version = job.executor.tpu_capability.tpu_runtime_version
      if i == 0 and job.executor.requirements.replicas > 1:
        raise ValueError('The first job in a JobGroup using the Caip executor '
                         'cannot have requirements.replicas > 1.')
      pool = aip_v1.WorkerPoolSpec(
          machine_spec=get_machine_spec(job),
          container_spec=aip_v1.ContainerSpec(
              image_uri=executable.image_path,
              args=args,
              env=env,
              # TODO: The `tpuTfVersion` field doesn't exist
              # yet in AI Platform (Unified). This is a placeholder
              # based on the legacy AI Platform ReplicaConfig.
              # https://cloud.google.com/ai-platform/training/docs/reference/rest/v1/projects.jobs#ReplicaConfig
              # 'tpuTfVersion': tpu_runtime_version,
          ),
          replica_count=job.executor.requirements.replicas,
      )
      # The model SDK only supports tensorboard in v1beta1.
      # https://github.com/googleapis/python-aiplatform/blob/master/google/cloud/aiplatform/jobs.py#L1201
      # But ContainerSpec.env is not supported in v1beta1.
      # https://github.com/googleapis/python-aiplatform/blob/master/google/cloud/aiplatform_v1beta1/types/custom_job.py#L216
      if tensorboard:
        if env:
          print('Enviornment variables are not supported when running with '
                'tensorboard. Set environment variables using `args` instead.')
        pool.container_spec.env = None
      pools.append(pool)

    # TOOD(chenandrew): CAIP only allows for 4 worker pools.
    # If CAIP does not implement more worker pools, another work-around
    # is to simply put all jobs into the same worker pool as replicas.
    # Each replica would contain the same image, but would execute different
    # modules based the replica index. A disadvantage of this implementation
    # is that every replica must have the same machine_spec.
    if len(pools) > 4:
      raise ValueError('Cloud Job for xm jobs {} contains {} worker types. '
                       'Only 4 worker types are supported'.format(
                           jobs, len(pools)))

    custom_job = aiplatform.CustomJob(
        project=self.project,
        location=self.location,
        display_name=name,
        worker_pool_specs=pools,
        staging_bucket=output_dir,
    )
    custom_job.run(
        sync=False,
        service_account=auth.get_service_account(),
        tensorboard=tensorboard,
    )
    custom_job.wait_for_resource_creation()
    print(f'Job launched at: {custom_job._dashboard_uri()}')  # pylint: disable=protected-access
    return custom_job.resource_name

  def get_tensorboard_settings(self, jobs: Sequence[xm.Job]) -> Tuple[str, str]:
    """Get the tensorboard settings for a sequence of Jobs."""
    executors = []
    for job in jobs:
      assert isinstance(job.executor, local_executors.Caip)
      executors.append(job.executor)
    if all(not executor.tensorboard for executor in executors):
      return '', ''

    if not executors[0].tensorboard:
      raise ValueError(
          'Jobs in this job group must have the same tensorboard settings. ' +
          'jobs[0] has no tensorboard settings.')
    output_dir = executors[0].tensorboard.base_output_directory
    tensorboard = executors[0].tensorboard.name
    for i, executor in enumerate(executors):
      if not executor:
        raise ValueError(
            'Jobs in this job group must have the same tensorboard settings. ' +
            'jobs[i] has no tensorboard settings.')
      if (executor.tensorboard.name != tensorboard or
          executor.tensorboard.base_output_directory != output_dir):
        raise ValueError(
            'Jobs in this job group must have the same tensorboard settings. ' +
            f'jobs[0] has tensorboard = {tensorboard} and output_dir =' +
            f'{output_dir}. jobs[{i}] has tensorboard = ' +
            f'{executor.tensorboard.name} and output_dir = '
            f'{executor.tensorboard.base_output_directory}.')
    if output_dir and not output_dir.startswith('gs://'):
      output_dir = os.path.join('gs://', output_dir)
    return tensorboard, output_dir

  async def wait_for_job(self, job_name: str) -> None:
    job = aiplatform.CustomJob.get(job_name)
    loop = asyncio.get_event_loop()
    return await loop.run_in_executor(None, job._block_until_complete)  # pylint: disable=protected-access

  def cancel(self, job_name: str) -> None:
    aiplatform.CustomJob.get(job_name).cancel()

  async def create_tensorboard(self, name: str) -> str:
    """Create a CAIP tensorboard instance."""
    tensorboard_client = aip_v1beta.TensorboardServiceAsyncClient(
        client_options={
            'api_endpoint': f'{self.location}-aiplatform.googleapis.com'
        })
    tensorboard = aip_v1beta.Tensorboard(display_name=name)
    op = await tensorboard_client.create_tensorboard(
        aip_v1beta.CreateTensorboardRequest(
            parent=self.parent,
            tensorboard=tensorboard,
        ))
    return (await op.result()).name

  def get_state(self, job_name: str) -> aip_v1.JobState:
    return aiplatform.CustomJob.get(job_name).state


def get_machine_spec(job: xm.Job) -> Dict[str, Any]:
  """Get the GCP machine type that best matches the Job's requirements."""
  assert isinstance(job.executor, local_executors.Caip)
  requirements = job.executor.requirements
  machine_type = cpu_ram_to_machine_type(
      requirements.task_requirements.get(xm.ResourceType.CPU),
      requirements.task_requirements.get(xm.ResourceType.RAM))
  spec = {'machine_type': machine_type}
  for resource, value in requirements.task_requirements.items():
    accelerator_type = None
    if resource in xm.GpuType:
      accelerator_type = 'NVIDIA_TESLA_' + str(resource).upper()
    elif resource in xm.TpuType:
      accelerator_type = _CLOUD_TPU_ACCELERATOR_TYPES[resource]
    if accelerator_type:
      # TPU_V2 and TPU_V3 removed in
      # https://github.com/googleapis/python-aiplatform/commit/f3a3d03c8509dc49c24139155a572dacbe954f66
      # Hardcode their values until they are reintroduced.
      if accelerator_type == 'TPU_V2':
        spec['accelerator_type'] = 6
      elif accelerator_type == 'TPU_V3':
        spec['accelerator_type'] = 7
      else:
        spec['accelerator_type'] = aip_v1.AcceleratorType[accelerator_type]
      spec['accelerator_count'] = int(value)
  return spec


@attr.s(auto_attribs=True)
class CaipHandle(local_execution.ExecutionHandle):
  """A handle for referring to the launched container."""

  job_name: str

  async def wait(self) -> None:
    await client().wait_for_job(self.job_name)

  def stop(self) -> None:
    client().cancel(self.job_name)

  def get_status(self) -> local_status.LocalWorkUnitStatus:
    state = client().get_state(self.job_name)
    status = _STATE_TO_STATUS[state]
    return local_status.LocalWorkUnitStatus(status=status)


# Must act on all jobs with `local_executors.Caip` executor.
def launch(experiment_title: str, work_unit_name: str,
           job_group: xm.JobGroup) -> List[CaipHandle]:
  """Launch CAIP jobs in the job_group and return a handler."""
  jobs = xm.job_operators.collect_jobs_by_filter(job_group, _caip_job_predicate)
  # As client creation may throw, do not initiate it if there are no jobs.
  if not jobs:
    return []

  job_name = client().launch(
      name=f'{experiment_title}_{work_unit_name}',
      jobs=jobs,
  )
  return [CaipHandle(job_name=job_name)]


def cpu_ram_to_machine_type(cpu: Optional[int], ram: Optional[int]) -> str:
  """Convert a cpu and memory spec into a machine type."""
  if cpu is None or ram is None:
    return 'n1-standard-4'

  optimal_machine_type = ''
  optimal_excess_resources = math.inf
  for machine_type, (machine_cpu,
                     machine_ram) in _MACHINE_TYPE_TO_CPU_RAM.items():
    if machine_cpu >= cpu and machine_ram >= ram:
      excess = machine_cpu + machine_ram - cpu - ram
      if excess < optimal_excess_resources:
        optimal_machine_type = machine_type
        optimal_excess_resources = excess

  if optimal_machine_type:
    return optimal_machine_type
  raise ValueError(
      '(cpu={}, ram={}) does not fit in any valid machine type'.format(
          cpu, ram))



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A collection of tools for files operations."""

import os
import tempfile


class TemporaryFilePath:
  """A context manager providing a temporary file path.

  Unlike NamedTemporaryFile, TemporaryFilePath closes the file when one enters
  the context.
  """

  _path: str

  def __enter__(self):
    fd, path = tempfile.mkstemp()
    os.close(fd)
    self._path = path
    return path

  def __exit__(self, error_type, error_value, traceback):
    os.remove(self._path)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for launching Vizier Explorations using Vertex Vizier."""

import abc
import asyncio
from typing import Any, Dict, Optional

from google.cloud import aiplatform_v1beta1 as aip

from xmanager import xm
from xmanager.cloud import auth
from xmanager.vizier.vizier_controller import VizierController

_DEFAULT_LOCATION = 'us-central1'


class VizierStudyFactory(abc.ABC):
  """Abstract class representing vizier study generator."""

  def __init__(self, location: str) -> None:
    self.vz_client = aip.VizierServiceClient(
        client_options=dict(
            api_endpoint=f'{location}-aiplatform.googleapis.com'))

  @abc.abstractmethod
  def study(self, experiment: xm.Experiment) -> aip.Study:
    """Create or load the study for the given `experiment`."""


class NewStudy(VizierStudyFactory):
  """Vizier study generator that generates new study from given config."""

  def __init__(self,
               study_spec: aip.StudySpec,
               display_name: Optional[str] = None,
               project: str = auth.get_project_name(),
               location: str = _DEFAULT_LOCATION) -> None:
    super().__init__(location)

    self._display_name = display_name
    self._study_spec = study_spec
    self._project = project
    self._location = location

  def study(self, experiment: xm.Experiment) -> aip.Study:
    return self.vz_client.create_study(
        parent=f'projects/{self._project}/locations/{self._location}',
        study=aip.Study(
            display_name=self._display_name or f'X{experiment.experiment_id}',
            study_spec=self._study_spec))


class VizierExploration:
  """An API for launching experiment as a Vizier-based Exploration."""

  def __init__(self, experiment: xm.Experiment, job: xm.JobType,
               study_factory: VizierStudyFactory, num_trials_total: int,
               num_parallel_trial_runs: int) -> None:
    """Create a VizierExploration.

    Args:
      experiment: the experiment who does the exploration.
      job: a job to run.
      study_factory: the VizierStudyFactory used to create or load the study.
      num_trials_total: total number of trials the experiment want to explore.
      num_parallel_trial_runs: number of parallel runs evaluating the trials.
    """

    # TODO: Reconsider to make functions async instead of
    # using get_event_loop().
    def work_unit_generator(vizier_params: Dict[str, Any]) -> xm.WorkUnit:
      return asyncio.get_event_loop().run_until_complete(
          experiment.add(job, self._to_job_params(vizier_params)))

    self._controller = VizierController(work_unit_generator,
                                        study_factory.vz_client,
                                        study_factory.study(experiment),
                                        num_trials_total,
                                        num_parallel_trial_runs)

  def _to_job_params(self, vizier_params: Dict[str, Any]) -> Dict[str, Any]:
    # TODO: unflatten parameters for JobGroup case (currently this
    # works for xm.Job).
    # For example: transform
    # {'learner.args.learning_rate': 0.1}
    # to
    # {'learner': {'args': {'learning_rate': 0.1}}}
    return {'args': vizier_params}

  def launch(self, **kwargs) -> None:
    self._controller.run(**kwargs)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""An interface to manipulate and access experiment metadata.

Metadata is attached to a context and the context may belong to an experiment
or a work unit.
"""

from typing import Collection, List, Optional, Set
import attr


class ContextAnnotations:
  """Interface for managing annotations.

  Annotations are user-supplied attributes of a context, such as title or tags.
  Default method implementations are intentionally left blank so that backends
  only have to implement the subset they support.
  """

  @property
  def title(self) -> str:
    """An experiment title.

    To differentiate experiments from each other they can be given a human
    readable title. Same title can be reused for multiple experiments.
    """
    return ''

  def set_title(self, title: str) -> None:
    """Sets the context title."""


@attr.s(auto_attribs=True)
class MetadataContext:
  """Interface for managing metadata.

  The metadata context could be attached to an experiment or a work unit.


  Attributes:
    creator: The username of the creator of this context.
    annotations: User-modifiable annotations.
  """

  creator: str
  annotations: ContextAnnotations



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Abstract API specification for XManager implementations.

Each implementation of the XManager API should override the abstract methods.
Users are normally expected to have the following pair of imports:

```
from xmanager import xm
from xmanager import xm_foo
```
"""

import abc
import asyncio
from concurrent import futures
import functools
import getpass
import inspect
import queue
import threading
from typing import Any, Awaitable, Callable, Collection, Dict, Mapping, Optional, Sequence, overload

import attr
import immutabledict
from xmanager.xm import async_packager
from xmanager.xm import id_predictor
from xmanager.xm import job_blocks
from xmanager.xm import job_operators
from xmanager.xm import metadata_context
from xmanager.xm import pattern_matching


def _check_if_unsupported_args_are_present(args: Mapping[str, Any],
                                           supported_args: Collection[str],
                                           job_type: str) -> None:
  supported_args = set(supported_args)
  unsupported_args = set(args.keys()) - supported_args
  if unsupported_args:
    raise ValueError(
        f'Arguments {unsupported_args!r} are not supported by {job_type}. Only '
        f'{supported_args!r} are allowed.')


def _apply_args_to_job(job: job_blocks.Job, args: Mapping[str, Any]) -> None:
  """Overrides job properties."""
  _check_if_unsupported_args_are_present(args, ('args', 'env_vars'), 'xm.Job')

  if 'args' in args:
    job.args = job_blocks.merge_args(job.args, args['args'])
  job.env_vars.update(args.get('env_vars', {}))


def _apply_args_to_job_group(job_group: job_blocks.JobGroup,
                             args: Mapping[str, Any]) -> None:
  """Recursively overrides job group properties."""
  if args:
    _check_if_unsupported_args_are_present(args, job_group.jobs.keys(),
                                           'xm.JobGroup')
    for key, job in job_group.jobs.items():
      _apply_args(job, args.get(key, {}))


_apply_args = pattern_matching.match(
    _apply_args_to_job, _apply_args_to_job_group,
    pattern_matching.Case([job_blocks.JobGeneratorType, Any],
                          lambda other, args: None))


class ExperimentUnitStatus(abc.ABC):
  """The status of an experiment unit."""

  @property
  @abc.abstractmethod
  def is_active(self) -> bool:
    """Returns whether the unit is not in terminal state.

    It may be actively running or queued. The unit may produce more results.
    If the unit is stopped by a user it will be neither active, completed
    nor failed.
    """
    raise NotImplementedError

  @property
  @abc.abstractmethod
  def is_completed(self) -> bool:
    """Returns whether the unit has completed without failures.

    This is a terminal state. The unit has produced all the intended results.
    But it still may be restarted by an explicit request.
    """
    raise NotImplementedError

  @property
  @abc.abstractmethod
  def is_failed(self) -> bool:
    """Returns whether the unit has failed.

    This is a terminal state. Experiment unit will enter this state on any
    fatal failure, such as process exiting with non-zero code, cloud rejecting
    to schedule/queue the job or exceptions in JobGenerator. The unit will stay
    in this state unless explicitly restarted.
    Intermediate failures do not result in this state.
    """
    raise NotImplementedError

  @property
  @abc.abstractmethod
  def message(self) -> str:
    """An optional human-readable message providing context for the status.

    This may take the form of explaining why the work unit is in this state,
    or any potentially transient errors the work unit may be experiencing.
    """
    raise NotImplementedError


class ExperimentUnitError(RuntimeError):
  """Experiment unit could not be completed."""


class ExperimentUnitFailedError(ExperimentUnitError):
  """A job running in an experiment unit has failed."""


class ExperimentUnitNotCompletedError(ExperimentUnitError):
  """Experiment unit is neither running nor completed.

  For example it may be stopped by a user.
  """


class NotFoundError(KeyError):
  """Experiment/Work Unit/etc. has not been found."""


def _work_unit_arguments(
    job: job_blocks.JobType,
    args: Mapping[str, Any],
) -> Mapping[str, Any]:
  """Constructs work unit arguments to display them in various UIs.

  If users pass `args` to the `.add` method explicitly, we assume `args` to be
  the sought work unit arguments. If `args` are not passed to `.add`, we deduce
  work unit arguments implicitly from the `job`s' `args` and `env_vars`.

  Args:
    job: A job to run inside a work unit.
    args: Explicitly specified arguments (could be empty).

  Returns:
    Depending on the type of the `job` given, can be one of the following:

      - if it's an instance of `Job`, we return `{'args': job.args, 'env_vars':
        job.env_vars}` with empty values omitted;
      - if it's an instance of `JobGroup`, we recursively unwind the group while
        populating corresponding nested dictionaries until we reach standalone
        `Job`s;
      - if it's a job generator, we return `{}`.
  """
  if args:
    # In order to give users control on what is shown as work unit arguments we
    # don't alter them if a value is given.
    return args

  def deduce_args_for_job(job: job_blocks.Job) -> Dict[str, Any]:
    args = {
        'args': job.args.to_dict(kwargs_only=True),
        'env_vars': job.env_vars
    }
    return {key: value for key, value in args.items() if value}

  def deduce_args_for_job_group(group: job_blocks.JobGroup) -> Dict[str, Any]:
    args = {}
    for job_name, job in group.jobs.items():
      job_args = deduce_args(job)
      if job_args:
        args[job_name] = job_args
    return args

  deduce_args = pattern_matching.match(
      deduce_args_for_job, deduce_args_for_job_group,
      pattern_matching.Case([job_blocks.JobGeneratorType],
                            lambda generator: {}))

  return deduce_args(job)


class ExperimentUnitRole(abc.ABC):
  """The role of an experiment unit within the experiment structure."""
  pass


class ExperimentUnit(abc.ABC):
  """ExperimentUnit is a collection of semantically associated `Job`s."""

  def __init__(self, experiment: 'Experiment',
               create_task: Callable[[Awaitable[Any]], futures.Future],
               args: Mapping[str, Any], role: ExperimentUnitRole) -> None:
    """Initializes an `ExperimentUnit` instance.

    Args:
      experiment: An experiment this unit belongs to.
      create_task: A callback to register a new asynchronous task.
      args: Arguments to this experiment unit. Most commonly used to represent
        the hyperparameter sweep trial corresponding to a work unit.
      role: The role of this unit in the experiment structure.
    """
    self._experiment = experiment
    self._create_task = create_task
    self._args = args
    self._role = role

    self._launched_error = None

  @property
  def experiment_id(self) -> int:
    """Returns a unique ID assigned to the experiment."""
    return self._experiment.experiment_id

  @property
  @functools.lru_cache()
  def _launched_event(self) -> asyncio.Event:
    """Event to be set when the experiment unit is launched."""
    # In Python < 3.8 `Event` must be created within the event loop. We defer
    # the initialization because the constructor is not run in the loop. Note
    # that after migration to Python >= 3.8 this code can be moved to the
    # constructor.
    return asyncio.Event()

  def add(
      self,
      job: job_blocks.JobType,
      args: Mapping[str, Any] = immutabledict.immutabledict()
  ) -> Awaitable[None]:
    # pyformat: disable
    """Adds a Job / JobGroup to the experiment unit.

    Only one JobGroup can be added to an ExperimentUnit. This limitation may be
    lifted in future versions.

    Args:
      job: A job or job group to add.
      args: Keyword arguments to be passed to the job. For Job and JobGroup args
        are recursively expanded. For example,

        ```
        wu.add(
            JobGroup(agent=Job(...)),
            args={'agent': {'args': {'learning_rate': 0.1}}},
        )
        ```

        would update `args` field of a job `agent` in the group.

    Returns:
      An awaitable that would be fulfilled when the job is launched.
    """
    # pyformat: enable
    job = job_operators.shallow_copy_job_type(job)
    _apply_args(job, args)
    job_operators.populate_job_names(job)

    def launch_job(job: job_blocks.Job) -> Awaitable[None]:
      return self._launch_job_group(
          job_blocks.JobGroup(**{job.name: job}),
          _work_unit_arguments(job, self._args))

    def launch_job_group(group: job_blocks.JobGroup) -> Awaitable[None]:
      return self._launch_job_group(group,
                                    _work_unit_arguments(group, self._args))

    def launch_job_generator(
        job_generator: job_blocks.JobGeneratorType) -> Awaitable[None]:
      if not inspect.iscoroutinefunction(job_generator):
        raise ValueError(
            'Job generator must be an async function. Signature needs to be '
            '`async def job_generator(work_unit: xm.WorkUnit):`')
      return job_generator(self, **args)

    job_awaitable = pattern_matching.match(launch_job, launch_job_group,
                                           launch_job_generator)(
                                               job)

    async def launch():
      try:
        await job_awaitable
      except Exception as e:
        self._launched_error = ExperimentUnitError(e)
        raise
      finally:
        # Note that the current implementation reuses `_launched_event`: imagine
        # a `work_unit.add(generator)` call -- the generator makes a
        # `work_unit.add` call internally as well, touching `_launched_event`
        # twice in total. Same would apply for multiple external `work_unit.add`
        # calls, which is not allowed yet. `_launched_error` is susceptible to
        # that too.
        self._launched_event.set()

    return asyncio.wrap_future(self._create_task(launch()))

  async def wait_until_complete(self):
    """Waits until the unit is in a final state: completed/failed/stopped.

    Raises:
      ExperimentUnitError: Exception if the unit couldn't complete.
    """
    await self._launched_event.wait()
    if self._launched_error:
      raise self._launched_error
    await self._wait_until_complete()

  async def _launch_job_group(self, job_group: job_blocks.JobGroup,
                              args_view: Mapping[str, Any]) -> None:
    """Launches a given job group as part of the unit."""
    raise NotImplementedError

  async def _wait_until_complete(self) -> None:
    """Waits until the unit is in a final state: completed/failed/stopped.

    Child classes need to implement this method to support awaiting units.

    Unlike wait_until_complete this method asumes that unit has been fully
    created. This method is only invoked if somebody has requested to monitor
    unit.
    """
    raise NotImplementedError

  def stop(self) -> None:
    """Initiate the process to stop the unit from running.

    This method will synchronously make a request for the unit to stop.
    However, the method does not actually wait for the unit to be in a
    terminal state.

    Use self.wait_until_complete() after self.stop() to guarantee the unit
    is stopped.
    """
    raise NotImplementedError

  def get_status(self) -> ExperimentUnitStatus:
    """Gets the status of this unit."""
    raise NotImplementedError

  @property
  @abc.abstractmethod
  def experiment_unit_name(self) -> str:
    raise NotImplementedError

  def get_full_job_name(self, job_name: str) -> str:
    """Given `Job.name` constructs its full name.

    The primary use case is addressing containers -- full names serve as
    hostnames.

    Args:
        job_name: Short name of a job.

    Returns:
        Full name of the job.
    """
    return f'{self.experiment_unit_name}_{job_name}'

  @property
  def context(self) -> metadata_context.MetadataContext:
    """Returns metadata context for a unit."""
    return metadata_context.MetadataContext(
        creator=getpass.getuser(),
        annotations=metadata_context.ContextAnnotations())


@attr.s(auto_attribs=True)
class WorkUnitRole(ExperimentUnitRole):
  """An experiment unit with this role is a work unit.

  Work units contain jobs that are often run as trials as part of an
  experiment's hyper-parameter search. The status of a work unit is used to
  determine the status of the experiment.
  """


class WorkUnit(ExperimentUnit):
  """Work units are experiment units with the work unit role."""

  @property
  @abc.abstractmethod
  def work_unit_id(self) -> int:
    raise NotImplementedError


@attr.s(auto_attribs=True)
class AuxiliaryUnitRole(ExperimentUnitRole):
  """An experiment unit with this role is an auxiliary unit.

  Auxiliary units contain jobs that are not part of the trials of a
  hyper-parameter search. The status of an auxiliary unit is not used to
  determine the status of the experiment. e.g. Tensorboard

  Attributes:
      termination_delay_secs: How long to keep AUX unit running after experiment
        completion.
  """

  termination_delay_secs: int


class Experiment(abc.ABC):
  """Experiment contains a family of jobs run on the same snapshot of code.

  Experiment also implements the behavior of how to add and execute jobs.
  Attempting to add jobs that contain Executables with unsupported types will
  fail.
  """

  # An event loop in which job generators would be run.
  _event_loop: asyncio.AbstractEventLoop
  # A queue of background tasks that launch work units.
  _running_tasks: queue.Queue
  # Work unit ID predictor.
  _work_unit_id_predictor: id_predictor.Predictor
  # A class variable for batching packaging requests.
  _async_packager: async_packager.AsyncPackager

  @property
  def experiment_id(self) -> int:
    """Returns a unique ID assigned to the experiment."""
    raise NotImplementedError

  def _enter(self) -> None:
    """Initializes internal state on context manager enter."""
    self._running_tasks = queue.Queue()
    self._work_unit_id_predictor = id_predictor.Predictor(1 +
                                                          self.work_unit_count)

  def __enter__(self):
    if asyncio.get_event_loop().is_running():
      raise RuntimeError('When using Experiment from a coroutine please use '
                         '`async with` syntax')

    self._event_loop = asyncio.new_event_loop()
    asyncio.get_child_watcher().attach_loop(self._event_loop)
    self._event_loop_thread = threading.Thread(
        target=self._event_loop.run_forever, daemon=True)
    self._event_loop_thread.start()

    self._enter()
    return self

  def _wait_for_tasks(self):
    while not self._running_tasks.empty():
      self._running_tasks.get_nowait().result()

  def __exit__(self, exc_type, exc_value, traceback):
    self._wait_for_tasks()
    self._event_loop.call_soon_threadsafe(self._event_loop.stop)
    self._event_loop_thread.join()

  async def __aenter__(self):
    self._event_loop = asyncio.get_event_loop()
    self._enter()
    return self

  async def _await_for_tasks(self):
    while not self._running_tasks.empty():
      await asyncio.wrap_future(self._running_tasks.get_nowait())

  async def __aexit__(self, exc_type, exc_value, traceback):
    await self._await_for_tasks()

  @classmethod
  def package(
      cls, packageables: Sequence[job_blocks.Packageable] = ()
  ) -> Sequence[job_blocks.Executable]:
    """Packages executable specs into executables based on the executor specs.

    Builds all given executables specs in parallel. While calling package()
    multiple times is allowed, that would result in slow sequential build,
    even if invoked from concurrent threads.

    Args:
      packageables: A sequence of packageables to build.

    Returns:
      A sequence of packaging results. Order corresponds to the order of input
      packageables.
    """
    return cls._async_packager.package(packageables)

  @classmethod
  def package_async(
      cls,
      packageable: job_blocks.Packageable) -> Awaitable[job_blocks.Executable]:
    """Queues executable spec to be packaged into executable.

    If gathering all packageables for a single `package()` call is inconvenient,
    one may request packaging with `package_async` and later trigger the build
    for the whole batch with `package()`.

    Usage:

      if eval:
        eval_executable = experiment.package_async(xm.blaze_binary(...))
      if train:
        train_executable = experiment.package_async(xm.blaze_binary(...))

      experiment.package()  # Explicitly trigger packaging.

      jobs = {}
      if eval:
        jobs['eval'] = xm.job(await eval_executable, ...)
      if train:
        jobs['train'] = xm.job(await train_executable, ...)

    Args:
      packageable: Executable spec to package.

    Returns:
      An awaitable for the packaging result.
    """
    return cls._async_packager.add(packageable)

  @overload
  def add(self,
          job: job_blocks.JobType,
          args: Mapping[str, Any] = ...,
          role: WorkUnitRole = ...) -> Awaitable[WorkUnit]:
    ...

  @overload
  def add(self, job: job_blocks.JobType, args: Mapping[str, Any],
          role: ExperimentUnitRole) -> Awaitable[ExperimentUnit]:
    ...

  @overload
  def add(
      self,
      job: job_blocks.JobType,
      args: Mapping[str, Any] = ...,
      *,  # parameters after “*” are keyword-only parameters
      role: ExperimentUnitRole
  ) -> Awaitable[ExperimentUnit]:
    ...

  # The ExecutableUnit return type is determined by the role.
  def add(self, job, args=immutabledict.immutabledict(), role=WorkUnitRole()):
    # pyformat: disable
    """Adds a Job / JobGroup to the experiment.

    A new Experiment Unit is created to run the job.

    Args:
      job: A Job or JobGroup to add.
      args: Keyword arguments to be passed to the job. For Job and JobGroup args
        are recursively expanded. For example,

        ```
        wu.add(
            JobGroup(agent=Job(...)),
            args={'agent': {'args': {'learning_rate': 0.1}}},
        )
        ```

        would update `args` field of a job `agent` in the group.
      role: The role of this unit in the experiment structure.

    Returns:
      An awaitable that would be fulfilled when the job is launched.
    """
    # pyformat: enable
    experiment_unit = self._create_experiment_unit(args, role)

    async def launch():
      await experiment_unit.add(job, args)
      return experiment_unit

    return asyncio.wrap_future(self._create_task(launch()))

  @abc.abstractmethod
  def _create_experiment_unit(self, args: Mapping[str, Any],
                              role: ExperimentUnitRole) -> ExperimentUnit:
    raise NotImplementedError

  def _create_task(self, task: Awaitable[Any]) -> futures.Future:
    future = asyncio.run_coroutine_threadsafe(task, loop=self._event_loop)
    self._running_tasks.put_nowait(future)
    return future

  @property
  def work_unit_count(self) -> int:
    """Returns how many work units the experiment has."""
    raise NotImplementedError

  @abc.abstractmethod
  def work_units(self) -> Mapping[int, WorkUnit]:
    """Returns a mapping from work_unit_id to an instance of the work unit."""
    raise NotImplementedError

  @property
  def context(self) -> metadata_context.MetadataContext:
    """Returns metadata context for the experiment."""
    return metadata_context.MetadataContext(
        creator=getpass.getuser(),
        annotations=metadata_context.ContextAnnotations())


@abc.abstractmethod
def create_experiment(experiment_title: Optional[str] = None) -> Experiment:
  """Returns a concrete Experiment instance."""
  raise NotImplementedError


@abc.abstractmethod
def get_experiment(experiment_id: int) -> Experiment:
  """Returns an Experiment instance associated with this experiment id.

  Args:
    experiment_id: An ID of an experiment to get.

  Raises:
    NotFoundError: If experiment is not found.
  """
  raise NotImplementedError



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyformat: disable
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# pylint: skip-file
# source: xm_local/storage/data.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='xm_local/storage/data.proto',
  package='xmanager',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1bxm_local/storage/data.proto\x12\x08xmanager\"\x8a\x01\n\x03Job\x12#\n\x05local\x18\x01 \x01(\x0b\x32\x12.xmanager.LocalJobH\x00\x12\'\n\x04\x63\x61ip\x18\x02 \x01(\x0b\x32\x17.xmanager.AIPlatformJobH\x00\x12-\n\nkubernetes\x18\x03 \x01(\x0b\x32\x17.xmanager.KubernetesJobH\x00\x42\x06\n\x04kind\"7\n\x08LocalJob\x12\x0b\n\x03pid\x18\x01 \x01(\t\x12\x0b\n\x03\x63md\x18\x02 \x01(\t\x12\x11\n\ttimestamp\x18\x03 \x01(\x03\"&\n\rAIPlatformJob\x12\x15\n\rresource_name\x18\x01 \x01(\t\"4\n\rKubernetesJob\x12\x11\n\tnamespace\x18\x01 \x01(\t\x12\x10\n\x08job_name\x18\x02 \x01(\tb\x06proto3'
)




_JOB = _descriptor.Descriptor(
  name='Job',
  full_name='xmanager.Job',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='local', full_name='xmanager.Job.local', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='caip', full_name='xmanager.Job.caip', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='kubernetes', full_name='xmanager.Job.kubernetes', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='kind', full_name='xmanager.Job.kind',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=42,
  serialized_end=180,
)


_LOCALJOB = _descriptor.Descriptor(
  name='LocalJob',
  full_name='xmanager.LocalJob',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='pid', full_name='xmanager.LocalJob.pid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cmd', full_name='xmanager.LocalJob.cmd', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='xmanager.LocalJob.timestamp', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=182,
  serialized_end=237,
)


_AIPLATFORMJOB = _descriptor.Descriptor(
  name='AIPlatformJob',
  full_name='xmanager.AIPlatformJob',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource_name', full_name='xmanager.AIPlatformJob.resource_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=239,
  serialized_end=277,
)


_KUBERNETESJOB = _descriptor.Descriptor(
  name='KubernetesJob',
  full_name='xmanager.KubernetesJob',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='namespace', full_name='xmanager.KubernetesJob.namespace', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='job_name', full_name='xmanager.KubernetesJob.job_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=279,
  serialized_end=331,
)

_JOB.fields_by_name['local'].message_type = _LOCALJOB
_JOB.fields_by_name['caip'].message_type = _AIPLATFORMJOB
_JOB.fields_by_name['kubernetes'].message_type = _KUBERNETESJOB
_JOB.oneofs_by_name['kind'].fields.append(
  _JOB.fields_by_name['local'])
_JOB.fields_by_name['local'].containing_oneof = _JOB.oneofs_by_name['kind']
_JOB.oneofs_by_name['kind'].fields.append(
  _JOB.fields_by_name['caip'])
_JOB.fields_by_name['caip'].containing_oneof = _JOB.oneofs_by_name['kind']
_JOB.oneofs_by_name['kind'].fields.append(
  _JOB.fields_by_name['kubernetes'])
_JOB.fields_by_name['kubernetes'].containing_oneof = _JOB.oneofs_by_name['kind']
DESCRIPTOR.message_types_by_name['Job'] = _JOB
DESCRIPTOR.message_types_by_name['LocalJob'] = _LOCALJOB
DESCRIPTOR.message_types_by_name['AIPlatformJob'] = _AIPLATFORMJOB
DESCRIPTOR.message_types_by_name['KubernetesJob'] = _KUBERNETESJOB
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Job = _reflection.GeneratedProtocolMessageType('Job', (_message.Message,), {
  'DESCRIPTOR' : _JOB,
  '__module__' : 'xm_local.storage.data_pb2'
  # @@protoc_insertion_point(class_scope:xmanager.Job)
  })
_sym_db.RegisterMessage(Job)

LocalJob = _reflection.GeneratedProtocolMessageType('LocalJob', (_message.Message,), {
  'DESCRIPTOR' : _LOCALJOB,
  '__module__' : 'xm_local.storage.data_pb2'
  # @@protoc_insertion_point(class_scope:xmanager.LocalJob)
  })
_sym_db.RegisterMessage(LocalJob)

AIPlatformJob = _reflection.GeneratedProtocolMessageType('AIPlatformJob', (_message.Message,), {
  'DESCRIPTOR' : _AIPLATFORMJOB,
  '__module__' : 'xm_local.storage.data_pb2'
  # @@protoc_insertion_point(class_scope:xmanager.AIPlatformJob)
  })
_sym_db.RegisterMessage(AIPlatformJob)

KubernetesJob = _reflection.GeneratedProtocolMessageType('KubernetesJob', (_message.Message,), {
  'DESCRIPTOR' : _KUBERNETESJOB,
  '__module__' : 'xm_local.storage.data_pb2'
  # @@protoc_insertion_point(class_scope:xmanager.KubernetesJob)
  })
_sym_db.RegisterMessage(KubernetesJob)


# @@protoc_insertion_point(module_scope)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Builds images for XManager Docker executables."""

import os
import shutil
import tempfile
from typing import Dict, List, Optional

from absl import flags
from absl import logging
import docker
from docker.utils import utils as docker_utils
import requests

from xmanager import xm
from xmanager.cloud import auth
from xmanager.cloud import cloud_build
from xmanager.cloud import docker_lib
from xmanager.docker import docker_adapter
from xmanager.xm import utils

_BUILD_IMAGE_LOCALLY = flags.DEFINE_boolean(
    'xm_build_image_locally', True,
    'Use local Docker to build images instead of remote Google Cloud Build. '
    'This is usually a lot faster but requires docker to be installed.')
_USE_DOCKER_COMMAND = flags.DEFINE_boolean(
    'xm_use_docker_command', True,
    'Call "docker build" in a subprocess rather than using Python docker '
    'client library when building the docker image locally. This provies a '
    'much nicer output for interactive use.')
_SHOW_DOCKER_COMMAND_PROGRESS = flags.DEFINE_boolean(
    'xm_show_docker_command_progress', False,
    'Show container output during the "docker build".')
_WRAP_LATE_BINDINGS = flags.DEFINE_boolean(
    'xm_wrap_late_bindings', False,
    'Feature flag to wrap and unwrap late bindings for network addresses. '
    'ONLY works with PythonContainer with default instructions or simple '
    'instructions that do not modify the file directory. '
    'REQUIRES ./entrypoint.sh to be the ENTRYPOINT.')

# TODO: Find a master image than is compatible with every
# combination (TF, Torch, JAX) X (CPU, GPU, TPU).
_DEFAULT_BASE_IMAGE = 'gcr.io/deeplearning-platform-release/base-cu110'
_DOCKERFILE_TEMPLATE = """
FROM {base_image}

RUN if ! id 1000; then useradd -m -u 1000 clouduser; fi

{instructions}

COPY entrypoint.sh ./entrypoint.sh
RUN chown -R 1000:root ./entrypoint.sh && chmod -R 775 ./entrypoint.sh

{entrypoint}
"""

_ENTRYPOINT_TEMPLATE = """#!/bin/bash

if [[ ! -z "$KUBE_GOOGLE_CLOUD_TPU_ENDPOINTS" ]]; then
  # TPU is available; set up expected env vars.
  TPU_IP_AND_PORT="${{KUBE_GOOGLE_CLOUD_TPU_ENDPOINTS:7}}"
  TPU_ADDRESS="$(cut -d':' -f1 <<< "$TPU_IP_AND_PORT")"
  echo "TPU_ADDRESS is $TPU_ADDRESS"
  export TPU_ADDRESS
  # Sometimes TPUs are not ready yet when the job starts. Wait until they are.
  while ! nc -z $TPU_ADDRESS 8470 ; do sleep 5 ; done
  # Expected by PyTorch.
  export XRT_TPU_CONFIG="tpu_worker;0;$TPU_IP_AND_PORT"
fi

{cmds}
"""


def build(py_executable: xm.PythonContainer,
          args: xm.SequentialArgs,
          env_vars: Dict[str, str],
          image_name: Optional[str] = None,
          project: Optional[str] = None,
          bucket: Optional[str] = None) -> str:
  """Build a Docker image from a Python project.

  Args:
    py_executable: The PythonContainer to build.
    args: Args to pass to the image.
    env_vars: Environment variables to set in the image.
    image_name: The image name that will be assigned to the resulting image.
    project: The project to use if CloudBuild is used.
    bucket: The bucket to upload if CloudBuild is used.

  Returns:
    The name of the built image.
  """
  if not image_name:
    image_name = _get_image_name(py_executable)
  dockerfile = _create_dockerfile(py_executable, args, env_vars)
  entrypoint = _create_entrypoint(py_executable)
  dirname = os.path.basename(py_executable.path)
  python_path = py_executable.path

  with tempfile.TemporaryDirectory() as wrapped_directory:
    if _WRAP_LATE_BINDINGS.value:
      _wrap_late_bindings(wrapped_directory, python_path, dockerfile)
      python_path = wrapped_directory
      dockerfile = os.path.join(python_path, 'Dockerfile')

    with tempfile.TemporaryDirectory() as staging:
      docker_lib.prepare_directory(staging, python_path, dirname, entrypoint,
                                   dockerfile)
      return build_by_dockerfile(staging, os.path.join(staging, 'Dockerfile'),
                                 image_name, project, bucket)


def build_by_dockerfile(path: str,
                        dockerfile: str,
                        image_name: str,
                        project: Optional[str] = None,
                        bucket: Optional[str] = None):
  """Build a Docker image from a Docker directory.

  Args:
    path: The directory to use for the Docker build context.
    dockerfile: The path of Dockerfile.
    image_name: The name to set the built image to.
    project: The project to use if CloudBuild is used.
    bucket: The bucket to upload if CloudBuild is used.

  Returns:
    The name of the built image.
  """
  print('Building Docker image, please wait...')
  if _BUILD_IMAGE_LOCALLY.value:
    try:
      docker_client = docker.from_env()
      logging.info('Local docker: %s', docker_client.version())
    except docker.errors.DockerException as e:
      logging.info(e)
      print('Failed to initialize local docker.')
      print('Falling back to CloudBuild. See INFO log for details.')
    except requests.exceptions.ConnectionError as e:
      logging.info(e)
      if 'Permission denied' in str(e):
        print('Looks like there is a permission problem with docker. '
              'Did you install sudoless docker?')
      else:
        print('Failed to connect to local docker instance.')
      print('Falling back to CloudBuild. See INFO log for details.')
    else:
      # TODO: Improve out-of-disk space handling.
      return docker_lib.build_docker_image(
          image_name,
          path,
          dockerfile,
          use_docker_command=_USE_DOCKER_COMMAND.value,
          show_docker_command_progress=_SHOW_DOCKER_COMMAND_PROGRESS.value)

  # If Dockerfile is not a direct child of path, then create a temp directory
  # that contains both the contents of path and Dockerfile.
  with tempfile.TemporaryDirectory() as tempdir:
    if os.path.dirname(dockerfile) != path:
      new_path = os.path.join(tempdir, os.path.basename(path))
      shutil.copytree(path, new_path)
      shutil.copyfile(dockerfile, os.path.join(path, 'Dockerfile'))
      path = new_path

    cloud_build_client = cloud_build.Client(project=project, bucket=bucket)
    repository, _ = docker_utils.parse_repository_tag(image_name)
    upload_name = repository.split('/')[-1]
    cloud_build_client.build_docker_image(image_name, path, upload_name)
    docker_adapter.instance().pull_image(image_name)
    return image_name


def push(image: str) -> str:
  return docker_lib.push_docker_image(image)


def _get_image_name(py_executable: xm.PythonContainer) -> str:
  image_name = os.path.basename(py_executable.path)
  project_name = auth.get_project_name()
  tag = docker_lib.create_tag()
  return f'gcr.io/{project_name}/{image_name}:{tag}'


def _get_base_image(py_executable: xm.PythonContainer) -> str:
  if py_executable.base_image:
    return py_executable.base_image
  return _DEFAULT_BASE_IMAGE


def _create_instructions(py_executable: xm.PythonContainer,
                         env_vars: Dict[str, str]) -> str:
  """Create Docker instructions."""
  set_env_vars = [f'ENV {key}="{value}"' for key, value in env_vars.items()]
  if py_executable.docker_instructions:
    return '\n'.join(py_executable.docker_instructions + set_env_vars)

  directory = os.path.basename(py_executable.path)
  return '\n'.join(
      list(default_steps(directory, py_executable.use_deep_module)) +
      set_env_vars)


def default_steps(directory: str, use_deep_module: bool) -> List[str]:
  """Default commands to use in the Dockerfile."""
  workdir_setup_prefix = []
  workdir_setup_suffix = []
  project_dir = f'/{directory}'
  if use_deep_module:
    # Setting a top-level work dir allows using the Python code without
    # modifying import statements.
    workdir_setup_prefix = [
        'RUN mkdir /workdir',
        'WORKDIR /workdir',
    ]
    project_dir = f'/workdir/{directory}'
  else:
    workdir_setup_suffix = [
        f'WORKDIR {directory}',
    ]

  return workdir_setup_prefix + [
      # Without setting LANG, RDL ran into an UnicodeDecodeError, similar to
      # what is described at [1]. This seems to be good practice and not hurt so
      # we're just always setting it.
      # [1] https://github.com/spotDL/spotify-downloader/issues/279
      'ENV LANG=C.UTF-8',
      # Updating and installing on the same line causes cache-busting.
      # https://docs.docker.com/develop/develop-images/dockerfile_best-practices/#run
      'RUN apt-get update && apt-get install -y git netcat',
      'RUN python -m pip install --upgrade pip setuptools',
      f'COPY {directory}/requirements.txt {project_dir}/requirements.txt',
      f'RUN python -m pip install -r {directory}/requirements.txt',
      # It is best practice to copy the project directory as late as possible,
      # rather than at the beginning. This allows Docker to reuse cached layers.
      # If copying the project files were the first step, a tiny modification to
      # the source code will invalidate the cache.
      # https://docs.docker.com/develop/develop-images/dockerfile_best-practices/#add-or-copy
      f'COPY {directory}/ {project_dir}',
      # Changing ownwership of project_dir, so that both users: UID 1000
      # and root are the co-owner of it.
      f'RUN chown -R 1000:root {project_dir} && chmod -R 775 {project_dir}',
  ] + workdir_setup_suffix


def _create_dockerfile(
    py_executable: xm.PythonContainer,
    args: xm.SequentialArgs,
    env_vars: Dict[str, str],
) -> str:
  """Creates a Dockerfile from a project executable."""
  base_image = _get_base_image(py_executable)
  instructions = _create_instructions(py_executable, env_vars)
  entrypoint = _create_entrypoint_cmd(args)
  contents = _DOCKERFILE_TEMPLATE.format(
      base_image=base_image, instructions=instructions, entrypoint=entrypoint)
  print('Dockerfile:', contents, sep='\n')
  t = tempfile.NamedTemporaryFile(delete=False)
  with open(t.name, 'w') as f:
    f.write(contents)
  return t.name


def _get_entrypoint_commands(py_executable: xm.PythonContainer) -> str:
  """Given the executable, return entrypoint commands."""
  if isinstance(py_executable.entrypoint, xm.ModuleName):
    cmds = [f'python -m {py_executable.entrypoint.module_name}']
  elif isinstance(py_executable.entrypoint, xm.CommandList):
    # Commands specified by the user are passed unchanged.
    cmds = py_executable.entrypoint.commands
  else:
    raise ValueError('Unsupported entrypoint type {}'.format(
        type(py_executable.entrypoint)))
  cmds = '\n'.join(cmds)
  # Allow passing extra parameters to the commands.
  if not cmds.endswith(('$@', '"$@"')):
    cmds = cmds + ' "$@"'
  return cmds


def _create_entrypoint(py_executable: xm.PythonContainer) -> str:
  """Create a bash entrypoint based on the base image."""
  contents = _ENTRYPOINT_TEMPLATE.format(
      cmds=_get_entrypoint_commands(py_executable))

  t = tempfile.NamedTemporaryFile(delete=False)
  with open(t.name, 'w') as f:
    f.write(contents)
  return t.name


def _create_entrypoint_cmd(args: xm.SequentialArgs) -> str:
  """Create the entrypoint command with optional args."""
  entrypoint_args = ['./entrypoint.sh']
  entrypoint_args.extend(args.to_list(utils.ARG_ESCAPER))
  entrypoint = ', '.join([f'"{arg}"' for arg in entrypoint_args])
  return f'ENTRYPOINT [{entrypoint}]'


def _wrap_late_bindings(destination: str, path: str, dockerfile: str) -> None:
  """Create a new path and dockerfile to wrap/unwrap late-bindings.

  TODO: Rather than only working PythonContainer, this method can
  also work on PrebuiltContainers. We do this by inspecting the entrypoint by
  using `docker.APIClient().inspect_image()`.

  Late bindings are special formatted strings that are evaluated at runtime. The
  primary use for late-bindings is to find the address of other jobs in CAIP
  which is only known at runtime and cannot be statically defined.

  Args:
    destination: An empty destination to contain the new project path and the
      new dockerfile will be destination/Dockerfile. The current contents of
      destination will be deleted.
    path: The current project path to build.
    dockerfile: The current dockerfile path needed to build the project.
  """
  shutil.rmtree(destination)
  shutil.copytree(path, destination)

  root_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

  shutil.copyfile(
      os.path.join(root_dir, 'cloud', 'data', 'wrapped_entrypoint.sh'),
      os.path.join(destination, 'wrapped_entrypoint.sh'))
  shutil.copyfile(
      os.path.join(root_dir, 'cloud', 'utils.py'),
      os.path.join(destination, 'caip_utils.py'))
  shutil.copyfile(
      os.path.join(root_dir, 'vizier', 'vizier_worker.py'),
      os.path.join(destination, 'vizier_worker.py'))

  new_dockerfile = os.path.join(destination, 'Dockerfile')
  insert_instructions = [
      'RUN chmod +x ./wrapped_entrypoint.sh',
  ]
  with open(dockerfile) as f:
    contents = f.read()
  contents = contents.replace('ENTRYPOINT',
                              '\n'.join(insert_instructions + ['ENTRYPOINT']))
  contents = contents.replace('ENTRYPOINT ["./entrypoint.sh',
                              'ENTRYPOINT ["./wrapped_entrypoint.sh')
  with open(new_dockerfile, 'w') as f:
    f.write(contents)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle
from typing import Sequence
import unittest

from xmanager.xm import async_packager
from xmanager.xm import job_blocks
from xmanager.xm import utils


def _package_batch(
    packageables: Sequence[job_blocks.Packageable]
) -> Sequence[job_blocks.Executable]:
  return [
      job_blocks.Executable(name=packageable.executable_spec.name)
      for packageable in packageables
  ]


class _TestExecutableSpec(job_blocks.ExecutableSpec):

  def __init__(self, name: str) -> None:
    self._name = name

  @property
  def name(self) -> str:
    return self._name


def _make_packageable(name: str) -> job_blocks.Packageable:
  return job_blocks.Packageable(
      executable_spec=_TestExecutableSpec(name),
      executor_spec=job_blocks.ExecutorSpec())


class AsyncPackagerTest(unittest.TestCase):

  @utils.run_in_asyncio_loop
  async def test_async_packager_end_to_end(self):
    packager = async_packager.AsyncPackager(_package_batch)
    executable1 = packager.add(_make_packageable('1'))
    executable2 = packager.add(_make_packageable('2'))
    packager.package()
    self.assertEqual((await executable1).name, '1')
    self.assertEqual((await executable2).name, '2')

  @utils.run_in_asyncio_loop
  async def test_package_with_extra_packageables(self):
    packager = async_packager.AsyncPackager(_package_batch)
    async_executable = packager.add(_make_packageable('async'))
    [extra_executable] = packager.package((_make_packageable('extra'),))
    self.assertEqual((await async_executable).name, 'async')
    self.assertEqual(extra_executable.name, 'extra')

  @utils.run_in_asyncio_loop
  async def test_package_is_required(self):
    packager = async_packager.AsyncPackager(_package_batch)
    executable = packager.add(_make_packageable(''))
    with self.assertRaises(async_packager.PackageHasNotBeenCalledError):
      await executable

  def test_awaitable_is_picklable(self):
    packager = async_packager.AsyncPackager(_package_batch)
    executable = packager.add(_make_packageable(''))
    packager.package()
    executable_str = pickle.dumps(executable)

    # Wait for the executable in a separate event loop, which did not even exist
    # when we requested packaging.
    @utils.run_in_asyncio_loop
    async def wait_for_it():
      await pickle.loads(executable_str)

    wait_for_it()


if __name__ == '__main__':
  unittest.main()



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import unittest

from xmanager.xm import utils


async def make_me_a_sandwich() -> str:
  return 'sandwich'


class ResourceType(enum.Enum):
  MINERALS = 1
  VESPEN = 2


class UtilsTest(unittest.TestCase):

  @utils.run_in_asyncio_loop
  async def test_run_in_asyncio_loop(self):
    self.assertEqual(await make_me_a_sandwich(), 'sandwich')

  def test_run_in_asyncio_loop_returns_value(self):
    self.assertEqual(
        utils.run_in_asyncio_loop(make_me_a_sandwich)(), 'sandwich')

  def test_arg_escaper(self):
    self.assertEqual(utils.ARG_ESCAPER(1.0), '1.0')
    self.assertEqual(utils.ARG_ESCAPER('Jonny Droptable'), "'Jonny Droptable'")
    self.assertEqual(utils.ARG_ESCAPER(ResourceType.VESPEN), 'VESPEN')


if __name__ == '__main__':
  unittest.main()



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for packagables."""

import unittest

from xmanager.xm import executables
from xmanager.xm import job_blocks
from xmanager.xm import packagables
from xmanager.xm_local import executors


class PackagablesTest(unittest.TestCase):

  def test_minimal_executable_spec(self):
    expected = job_blocks.Packageable(
        executable_spec=executables.BazelBinary(label='label'),
        executor_spec=executors.Local.Spec(),
        args=[],
        env_vars={},
    )

    actual = packagables.bazel_binary(executors.Local.Spec(), label='label')

    self.assertEqual(actual, expected)

  def test_pkg_args_env_vars(self):
    expected = job_blocks.Packageable(
        executable_spec=executables.BazelBinary(label='label'),
        executor_spec=executors.Local.Spec(),
        args=['-f'],
        env_vars={'KEY': 'value'},
    )

    actual = packagables.bazel_binary(
        executors.Local.Spec(),
        label='label',
        args=['-f'],
        env_vars={'KEY': 'value'},
    )

    self.assertEqual(actual, expected)


if __name__ == '__main__':
  unittest.main()



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Run Job as a Vizier worker to manager WorkUnit Vizier interaction."""
import re
from typing import Dict, Optional

from absl import logging
from google.cloud import aiplatform_v1beta1 as aip

_TRIAL_NAME_REGEX = r'projects\/[^\/]+\/locations\/[^\/]+\/studies\/[^\/]+\/trials\/[^\/]+'


class VizierWorker:
  """Worker that manage interaction between job and Vizier."""

  def __init__(self, trial_name: str) -> None:
    if not re.match(_TRIAL_NAME_REGEX, trial_name):
      raise Exception('The trial_name must be in the form: '
                      'projects/{project}/locations/{location}/'
                      'studies/{study}/trials/{trial}')

    self._trial_name = trial_name

    location = trial_name.split('/')[3]
    self._vz_client = aip.VizierServiceClient(client_options={
        'api_endpoint': f'{location}-aiplatform.googleapis.com',
    })

  def add_trial_measurement(self, step: int, metrics: Dict[str, float]) -> None:
    """Add trial measurements to Vizier."""
    self._vz_client.add_trial_measurement(
        request=aip.AddTrialMeasurementRequest(
            trial_name=self._trial_name,
            measurement=aip.Measurement(
                step_count=step,
                metrics=[
                    aip.Measurement.Metric(metric_id=k, value=v)
                    for k, v in metrics.items()
                ],
            )))
    logging.info('Step %d Metric %s is reported', step, metrics)

  def complete_trial(self, infeasible_reason: Optional[str] = None) -> None:
    """Complete a trial."""
    self._vz_client.complete_trial(
        request=aip.CompleteTrialRequest(
            name=self._trial_name,
            trial_infeasible=infeasible_reason is not None,
            infeasible_reason=infeasible_reason))
    logging.info('Trial %s is completed', self._trial_name)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementation of local work unit statuses."""
import enum

from xmanager import xm


class LocalWorkUnitStatusEnum(enum.Enum):
  """Status of a local experiment job."""

  # Work unit was created, but has not terminated yet.
  RUNNING = 1
  # Work unit terminated and was successful.
  COMPLETED = 2
  # Work unit terminated and was not succesful.
  FAILED = 3
  # Work unit terminated because it was cancelled by the user.
  CANCELLED = 4


class LocalWorkUnitStatus(xm.ExperimentUnitStatus):
  """Status of a local experiment job."""

  def __init__(self,
               status: LocalWorkUnitStatusEnum,
               message: str = '') -> None:
    super().__init__()
    self._status = status
    self._message = message

  @property
  def is_active(self) -> bool:
    return self._status == LocalWorkUnitStatusEnum.RUNNING

  @property
  def is_completed(self) -> bool:
    return self._status == LocalWorkUnitStatusEnum.COMPLETED

  @property
  def is_failed(self) -> bool:
    return self._status == LocalWorkUnitStatusEnum.FAILED

  @property
  def message(self) -> str:
    return self._message



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for xmanager.cloud.kubernetes."""
import json
import unittest
from unittest import mock

from kubernetes import client as k8s_client

from xmanager import xm
from xmanager.cloud import kubernetes
from xmanager.xm_local import executables as local_executables
from xmanager.xm_local import executors as local_executors


class CallAPIResponse:
  items = []


class KubernetesTest(unittest.TestCase):

  def test_launch(self):
    fake_client = mock.Mock()
    fake_client.call_api.return_value = CallAPIResponse()
    client = kubernetes.Client(fake_client)

    job = xm.Job(
        name='test-job',
        executable=local_executables.GoogleContainerRegistryImage(
            name='test-image',
            image_path='image-path',
            args=xm.SequentialArgs.from_collection({'a': 1}),
        ),
        executor=local_executors.Kubernetes(
            xm.JobRequirements(cpu=1, ram=1, t4=2)),
        args={
            'b': 2,
            'c': 3
        },
    )
    expected_service = k8s_client.V1Service(
        metadata=k8s_client.V1ObjectMeta(name='experiments'),
        spec=k8s_client.V1ServiceSpec(
            selector={'service': 'experiments'},
            cluster_ip='None',
        ),
    )
    cluster_spec = json.dumps({
        'cluster': {
            'workerpool0': [
                'workerpool0.experiments.default.svc.cluster.local:2222'
            ]
        },
        'task': {
            'type': 'workerpool0',
            'index': 0
        }
    })
    expected_job = k8s_client.V1Job(
        metadata=k8s_client.V1ObjectMeta(name='test-job'),
        spec=k8s_client.V1JobSpec(
            template=k8s_client.V1PodTemplateSpec(
                metadata=k8s_client.V1ObjectMeta(
                    labels={'service': 'experiments'},
                    annotations={},
                ),
                spec=k8s_client.V1PodSpec(
                    hostname='workerpool0',
                    subdomain='experiments',
                    restart_policy='Never',
                    containers=[
                        k8s_client.V1Container(
                            name='test-job',
                            image='image-path',
                            resources=k8s_client.V1ResourceRequirements(
                                limits={
                                    'cpu': '1',
                                    'memory': '1',
                                    'nvidia.com/gpu': '2',
                                },),
                            args=['--a', '1', '--b', '2', '--c', '3'],
                            env=[
                                k8s_client.V1EnvVar(
                                    'CLUSTER_SPEC',
                                    cluster_spec,
                                )
                            ],
                        )
                    ],
                    node_selector={
                        'cloud.google.com/gke-accelerator': 'nvidia-tesla-t4',
                    },
                ),
            ),
            backoff_limit=0,
        ),
    )

    client.launch(lambda x: x, [job])
    [_, service_call, job_call] = fake_client.call_api.call_args_list
    _, service_kwargs = service_call
    self.assertEqual(service_kwargs['body'], expected_service)
    _, job_kwargs = job_call
    self.assertEqual(job_kwargs['body'], expected_job)

  def test_requirements_from_executor(self):
    executor = local_executors.Kubernetes(
        requirements=xm.JobRequirements(cpu=1, ram=1 * xm.GiB))
    requirements = kubernetes.requirements_from_executor(executor).to_dict()
    self.assertDictEqual(requirements['limits'], {
        'cpu': '1',
        'memory': str(2**30),
    })

  def test_requirements_from_executor_gpu(self):
    executor = local_executors.Kubernetes(
        requirements=xm.JobRequirements(v100=4))
    requirements = kubernetes.requirements_from_executor(executor).to_dict()
    self.assertDictEqual(requirements['limits'], {'nvidia.com/gpu': '4'})

  def test_requirements_from_executor_empty(self):
    executor = local_executors.Kubernetes()
    requirements = kubernetes.requirements_from_executor(executor).to_dict()
    self.assertDictEqual(requirements['limits'], {})

  def test_annotations_from_executor_tpu(self):
    executor = local_executors.Kubernetes(xm.JobRequirements(tpu_v2=8))
    self.assertDictEqual(
        kubernetes.annotations_from_executor(executor),
        {'tf-version.cloud-tpus.google.com': 'nightly'})

  def test_annotations_from_executor_gpu(self):
    executor = local_executors.Kubernetes(xm.JobRequirements(v100=4))
    self.assertDictEqual(kubernetes.annotations_from_executor(executor), {})

  def test_node_selector_from_executor_gpu(self):
    executor = local_executors.Kubernetes(xm.JobRequirements(v100=4))
    self.assertDictEqual(
        kubernetes.node_selector_from_executor(executor),
        {'cloud.google.com/gke-accelerator': 'nvidia-tesla-v100'})

  def test_node_selector_from_executor_tpu(self):
    executor = local_executors.Kubernetes(xm.JobRequirements(tpu_v2=8))
    self.assertDictEqual(kubernetes.node_selector_from_executor(executor), {})

  def test_node_selector_from_executor_empty(self):
    executor = local_executors.Kubernetes(xm.JobRequirements())
    self.assertDictEqual(kubernetes.node_selector_from_executor(executor), {})


if __name__ == '__main__':
  unittest.main()
